# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import os

import galsim
from galsim_test_helpers import *

imgdir = os.path.join(".", "SBProfile_comparison_images") # Directory containing the reference
                                                          # images.

@timer
def test_add():
    """Test the addition of two rescaled Gaussian profiles against a known double Gaussian result.
    """
    savedImg = galsim.fits.read(os.path.join(imgdir, "double_gaussian.fits"))
    dx = 0.2
    myImg = galsim.ImageF(savedImg.bounds, scale=dx)
    myImg.setCenter(0,0)

    gauss1 = galsim.Gaussian(flux=0.75, sigma=1)
    gauss2 = galsim.Gaussian(flux=0.25, sigma=3)
    sum_gauss = galsim.Add(gauss1,gauss2)
    sum_gauss.drawImage(myImg,scale=dx, method="sb", use_true_center=False)
    printval(myImg, savedImg)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Add(gauss1,gauss2) disagrees with expected result")

    cen = galsim.PositionD(0,0)
    np.testing.assert_equal(sum_gauss.centroid, cen)
    np.testing.assert_almost_equal(sum_gauss.flux, gauss1.flux + gauss2.flux)
    np.testing.assert_almost_equal(sum_gauss.xValue(cen), sum_gauss.max_sb)

    # Check with default_params
    sum_gauss = galsim.Add(gauss1,gauss2,gsparams=default_params)
    sum_gauss.drawImage(myImg,scale=dx, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Add(gauss1,gauss2) with default_params disagrees with "
            "expected result")
    sum_gauss = galsim.Add(gauss1,gauss2,gsparams=galsim.GSParams())
    sum_gauss.drawImage(myImg,scale=dx, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Add(gauss1,gauss2) with GSParams() disagrees with "
            "expected result")

    # Other ways to do the sum:
    sum_gauss = gauss1 + gauss2
    sum_gauss.drawImage(myImg,scale=dx, method="sb", use_true_center=False)
    printval(myImg, savedImg)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject gauss1 + gauss2 disagrees with expected result")
    sum_gauss = gauss1
    sum_gauss += gauss2
    sum_gauss.drawImage(myImg,scale=dx, method="sb", use_true_center=False)
    printval(myImg, savedImg)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject sum = gauss1; sum += gauss2 disagrees with expected result")
    sum_gauss = galsim.Add([gauss1,gauss2])
    sum_gauss.drawImage(myImg,scale=dx, method="sb", use_true_center=False)
    printval(myImg, savedImg)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Add([gauss1,gauss2]) disagrees with expected result")
    gauss1 = galsim.Gaussian(flux=1, sigma=1)
    gauss2 = galsim.Gaussian(flux=1, sigma=3)
    sum_gauss = 0.75 * gauss1 + 0.25 * gauss2
    sum_gauss.drawImage(myImg,scale=dx, method="sb", use_true_center=False)
    printval(myImg, savedImg)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject 0.75 * gauss1 + 0.25 * gauss2 disagrees with expected result")
    sum_gauss = 0.75 * gauss1
    sum_gauss += 0.25 * gauss2
    sum_gauss.drawImage(myImg,scale=dx, method="sb", use_true_center=False)
    printval(myImg, savedImg)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject sum += 0.25 * gauss2 disagrees with expected result")

    check_basic(sum_gauss, "sum of 2 Gaussians")

    # Test photon shooting.
    do_shoot(sum_gauss,myImg,"sum of 2 Gaussians")

    # Test kvalues
    do_kvalue(sum_gauss,myImg,"sum of 2 Gaussians")

    # Check picklability
    check_pickle(sum_gauss, lambda x: x.drawImage(method='sb'))
    check_pickle(sum_gauss)

    # Sum of just one argument should be equivalent to that argument.
    single = galsim.Add(gauss1)
    gsobject_compare(single, gauss1)
    check_basic(single, "`sum' of 1 Gaussian")
    check_pickle(single)
    do_shoot(single, myImg, "Single Sum")

    single = galsim.Add([gauss1])
    gsobject_compare(single, gauss1)
    check_basic(single, "`sum' of 1 Gaussian")
    check_pickle(single)

    # Should raise an exception for invalid arguments
    assert_raises(TypeError, galsim.Add)
    assert_raises(TypeError, galsim.Add, myImg)
    assert_raises(TypeError, galsim.Add, [myImg])
    assert_raises(TypeError, galsim.Add, [gauss1, myImg])
    assert_raises(TypeError, galsim.Add, [gauss1, gauss1, myImg])
    assert_raises(TypeError, galsim.Add, [gauss1, gauss1], real_space=False)
    assert_raises(TypeError, galsim.Sum)
    assert_raises(TypeError, galsim.Sum, myImg)
    assert_raises(TypeError, galsim.Sum, [myImg])
    assert_raises(TypeError, galsim.Sum, [gauss1, myImg])
    assert_raises(TypeError, galsim.Sum, [gauss1, gauss1, myImg])
    assert_raises(TypeError, galsim.Sum, [gauss1, gauss1], real_space=False)


@timer
def test_sub_neg():
    """Test that a - b is the same as a + (-b)."""
    a = galsim.Gaussian(fwhm=1)
    b = galsim.Kolmogorov(fwhm=1)

    c = a - b
    d = a + (-b)

    assert c == d

    im1 = c.drawImage()
    im2 = d.drawImage(im1.copy())

    np.testing.assert_equal(im1.array, im2.array)

@timer
def test_add_flux_scaling():
    """Test flux scaling for Add.
    """
    # decimal point to go to for parameter value comparisons
    param_decimal = 12
    test_flux = 17.9
    test_sigma = 1.8
    test_scale = 1.9

    # init with Gaussian and Exponential only (should be ok given last tests)
    obj = galsim.Add([galsim.Gaussian(sigma=test_sigma, flux=test_flux * .5),
                      galsim.Exponential(scale_radius=test_scale, flux=test_flux * .5)])
    obj *= 2.
    np.testing.assert_almost_equal(
        obj.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __imul__.")
    obj = galsim.Add([galsim.Gaussian(sigma=test_sigma, flux=test_flux * .5),
                      galsim.Exponential(scale_radius=test_scale, flux=test_flux * .5)])
    obj /= 2.
    np.testing.assert_almost_equal(
        obj.flux, test_flux / 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __idiv__.")
    obj = galsim.Add([galsim.Gaussian(sigma=test_sigma, flux=test_flux * .5),
                      galsim.Exponential(scale_radius=test_scale, flux=test_flux * .5)])
    obj2 = obj * 2.
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __rmul__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __rmul__ (result).")
    obj = galsim.Add([galsim.Gaussian(sigma=test_sigma, flux=test_flux * .5),
                      galsim.Exponential(scale_radius=test_scale, flux=test_flux * .5)])
    obj2 = 2. * obj
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __mul__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __mul__ (result).")
    obj = galsim.Add([galsim.Gaussian(sigma=test_sigma, flux=test_flux * .5),
                      galsim.Exponential(scale_radius=test_scale, flux=test_flux * .5)])
    obj2 = obj / 2.
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __div__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux / 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __div__ (result).")


@timer
def test_ne():
    """ Check that inequality works as expected."""
    gsp = galsim.GSParams(maxk_threshold=1.1e-3, folding_threshold=5.1e-3)
    gal1 = galsim.Gaussian(fwhm=1)
    gal2 = galsim.Gaussian(fwhm=2)

    # Sum.  Params are objs to add and potentially gsparams.
    gals = [galsim.Sum(gal1),
            galsim.Sum(gal1, gal2),
            galsim.Sum(gal2, gal1),  # Not! commutative.  (but is associative)
            galsim.Sum(galsim.Sum(gal1, gal2), gal2),
            galsim.Sum(gal1, gsparams=gsp),
            galsim.Sum(gal1, gsparams=gsp, propagate_gsparams=False)]
    check_all_diff(gals)


@timer
def test_sum_transform():
    """This test addresses a bug found by Ismael Serrano, #763, wherein some attributes
    got messed up for a Transform(Sum(Transform())) object.

    The bug was that we didn't bother to make a new SBProfile for a Sum (or Convolve) of
    a single object.  But if that was an SBTransform, then the next Transform operation
    combined the two Transforms, which messed up the repr.

    The fix is to always make an SBAdd or SBConvolve object even if the list of things to add
    or convolve only has one element.
    """
    gal0 = galsim.Exponential(scale_radius=0.34, flux=105.).shear(g1=-0.56,g2=0.15)

    for gal1 in [ galsim.Sum(gal0), galsim.Convolve(gal0) ]:
        gal2 = gal1.dilate(1)

        sgal1 = eval(str(gal1))
        rgal1 = eval(repr(gal1))
        sgal2 = eval(str(gal2))
        rgal2 = eval(repr(gal2))

        gal1_im = gal1.drawImage(nx=64, ny=64, scale=0.2)
        sgal1_im = sgal1.drawImage(nx=64, ny=64, scale=0.2)
        rgal1_im = rgal1.drawImage(nx=64, ny=64, scale=0.2)

        gal2_im = gal2.drawImage(nx=64, ny=64, scale=0.2)
        sgal2_im = sgal2.drawImage(nx=64, ny=64, scale=0.2)
        rgal2_im = rgal2.drawImage(nx=64, ny=64, scale=0.2)

        # Check that the objects are equivalent, even if they may be written differently.
        np.testing.assert_almost_equal(gal1_im.array, sgal1_im.array, decimal=8)
        np.testing.assert_almost_equal(gal1_im.array, rgal1_im.array, decimal=8)

        # These two used to fail.
        np.testing.assert_almost_equal(gal2_im.array, sgal2_im.array, decimal=8)
        np.testing.assert_almost_equal(gal2_im.array, rgal2_im.array, decimal=8)

        check_pickle(gal0)
        check_pickle(gal1)
        check_pickle(gal2)

@timer
def test_sum_noise():
    """Test that noise propagation works properly for compound objects.
    """
    obj1 = galsim.Gaussian(sigma=1.7)
    obj2 = galsim.Gaussian(sigma=2.3)
    obj1.noise = galsim.UncorrelatedNoise(variance=0.3, scale=0.2)
    obj2.noise = galsim.UncorrelatedNoise(variance=0.5, scale=0.2)
    obj3 = galsim.Gaussian(sigma=2.9)

    # Sum adds the variance of the components
    sum2 = galsim.Sum([obj1,obj2])
    np.testing.assert_almost_equal(sum2.noise.getVariance(), 0.8,
            err_msg = "Sum of two objects did not add noise varinace")
    sum2 = galsim.Sum([obj1,obj3])
    np.testing.assert_almost_equal(sum2.noise.getVariance(), 0.3,
            err_msg = "Sum of two objects did not add noise varinace")
    sum2 = galsim.Sum([obj2,obj3])
    np.testing.assert_almost_equal(sum2.noise.getVariance(), 0.5,
            err_msg = "Sum of two objects did not add noise varinace")
    sum3 = galsim.Sum([obj1,obj2,obj3])
    np.testing.assert_almost_equal(sum3.noise.getVariance(), 0.8,
            err_msg = "Sum of three objects did not add noise varinace")
    sum3 = obj1 + obj2 + obj3
    np.testing.assert_almost_equal(sum3.noise.getVariance(), 0.8,
            err_msg = "Sum of three objects did not add noise varinace")

    # Adding noise objects with different WCSs will raise a warning.
    obj4 = galsim.Gaussian(sigma=3.3)
    obj4.noise = galsim.UncorrelatedNoise(variance=0.3, scale=0.8)
    try:
        np.testing.assert_warns(galsim.GalSimWarning, galsim.Sum, [obj1, obj4])
    except:
        pass

@timer
def test_gsparams():
    """Test withGSParams with some non-default gsparams
    """
    obj1 = galsim.Exponential(half_light_radius=1.7)
    obj2 = galsim.Pixel(scale=0.2)
    gsp = galsim.GSParams(folding_threshold=1.e-4, maxk_threshold=1.e-4, maximum_fft_size=1.e4)
    gsp2 = galsim.GSParams(folding_threshold=1.e-2, maxk_threshold=1.e-2)

    sum = galsim.Sum(obj1, obj2)
    sum1 = sum.withGSParams(gsp)
    assert sum.gsparams == galsim.GSParams()
    assert sum1.gsparams == gsp
    assert sum1.obj_list[0].gsparams == gsp
    assert sum1.obj_list[1].gsparams == gsp

    sum2 = galsim.Sum(obj1.withGSParams(gsp), obj2.withGSParams(gsp))
    sum3 = galsim.Sum(galsim.Exponential(half_light_radius=1.7, gsparams=gsp),
                       galsim.Pixel(scale=0.2))
    sum4 = galsim.Add(obj1, obj2, gsparams=gsp)
    assert sum != sum1
    assert sum1 == sum2
    assert sum1 == sum3
    assert sum1 == sum4
    print('stepk = ',sum.stepk, sum1.stepk)
    assert sum1.stepk < sum.stepk
    print('maxk = ',sum.maxk, sum1.maxk)
    assert sum1.maxk > sum.maxk

    sum5 = galsim.Add(obj1, obj2, gsparams=gsp, propagate_gsparams=False)
    assert sum5 != sum4
    assert sum5.gsparams == gsp
    assert sum5.obj_list[0].gsparams == galsim.GSParams()
    assert sum5.obj_list[1].gsparams == galsim.GSParams()

    sum6 = sum5.withGSParams(gsp2)
    assert sum6 != sum5
    assert sum6.gsparams == gsp2
    assert sum6.obj_list[0].gsparams == galsim.GSParams()
    assert sum6.obj_list[1].gsparams == galsim.GSParams()



if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'BaseDeviate', 'UniformDeviate', 'GaussianDeviate', 'PoissonDeviate', 'DistDeviate',
            'BinomialDeviate', 'Chi2Deviate', 'GammaDeviate', 'WeibullDeviate', ]

import numpy as np
import os

from . import _galsim
from .errors import GalSimRangeError, GalSimValueError, GalSimIncompatibleValuesError
from .errors import galsim_warn
from ._utilities import isinteger, math_eval
from .table import LookupTable
from . import integ

class BaseDeviate:
    """Base class for all the various random deviates.

    This holds the essential random number generator that all the other classes use.

    All deviates take an initial ``seed`` argument that is used to seed the underlying random number
    generator.  It has three different kinds of behavior.

    1. An integer value can be provided to explicitly seed the random number generator with a
       particular value.  This is useful to have deterministic behavior.  If you seed with an
       integer value, the subsequent series of "random" values will be the same each time you
       run the program.

    2. A seed of 0 or None means to pick some arbitrary value that will be different each time
       you run the program.  Currently, this tries to get a seed from /dev/urandom if possible.
       If that doesn't work, then it creates a seed from the current time.  You can also get this
       behavior by omitting the seed argument entirely.  (i.e. the default is None.)

    3. Providing another BaseDeviate object as the seed will make the new BaseDeviate share the
       same underlying random number generator as the other BaseDeviate.  So you can make one
       BaseDeviate (of any type), and seed it with a particular deterministic value.  Then if you
       pass that BaseDeviate to any other one you make, they will both be using the same RNG and
       the series of "random" values will be deterministic.

    **Usage**:

    The only kind of random number you can obtain from a pure BaseDeviate (i.e. one that is
    not actually one of the variosu subclasses) is a "raw" value.  This is an unsigned 32-bit
    integer that behind the scenes is used by all sub-classes to generate floating point values
    with various distributions.

        >>> rng = galsim.BaseDeviate(215324)
        >>> rng.raw()
        3559052779

    Most other usage is effected by creating sub-classes corresponding to particular random
    deviates with various distributions.  E.g. `UniformDeviate` generates random values following
    a uniform distribution between 0 and 1.

        >>> rng = galsim.BaseDeviate(215324)
        >>> ud = galsim.UniformDeviate(rng)
        >>> ud()
        0.58736140513792634
        >>> ud2 = galsim.UniformDeviate(215324)
        >>> ud2()
        0.58736140513792634
    """
    def __init__(self, seed=None):
        self._rng_type = _galsim.BaseDeviateImpl
        self._rng_args = ()
        self.reset(seed)

    def seed(self, seed=0):
        """Seed the pseudo-random number generator with a given integer value.

        Parameters:
            seed:       An int value to be used to seed the random number generator.  Using 0
                        means to generate a seed from the system. [default: 0]
        """
        if seed == int(seed):
            self._seed(int(seed))
        else:
            raise TypeError("BaseDeviate seed must be an integer.  Got %s"%seed)

    def _seed(self, seed=0):
        """Equivalent to `seed`, but without any type checking.
        """
        self._rng.seed(seed)

    def reset(self, seed=None):
        """Reset the pseudo-random number generator, severing connections to any other deviates.
        Providing another `BaseDeviate` object as the seed connects this deviate with the other
        one, so they will both use the same underlying random number generator.

        Parameters:
            seed:       Something that can seed a `BaseDeviate`: an integer seed or another
                        `BaseDeviate`.  Using None means to generate a seed from the system.
                        [default: None]
        """
        if isinstance(seed, BaseDeviate):
            self._reset(seed)
        elif isinstance(seed, str):
            self._rng = self._rng_type(_galsim.BaseDeviateImpl(seed), *self._rng_args)
        elif seed is None:
            self._rng = self._rng_type(_galsim.BaseDeviateImpl(0), *self._rng_args)
        elif isinteger(seed):
            self._rng = self._rng_type(_galsim.BaseDeviateImpl(int(seed)), *self._rng_args)
        else:
            raise TypeError("BaseDeviate must be initialized with either an int or another "
                            "BaseDeviate")

    def _reset(self, rng):
        """Equivalent to `reset`, but rng must be a `BaseDeviate` (not an int), and there
        is no type checking.
        """
        self._rng = self._rng_type(rng._rng, *self._rng_args)

    @property
    def np(self):
        """Shorthand for self.as_numpy_generator()
        """
        return self.as_numpy_generator()

    def as_numpy_generator(self):
        """Return a numpy.random.Generator object that uses the current BaseDeviate for the
        underlying bit generations.

        This allows you to use the (probably) more familiar numpy functions, while maintaining
        GalSim's guarantees about random number stability across platforms.

        Example::

            >>> rng = galsim.BaseDeviate(1234)
            >>> gen = rng.as_numpy_generator()
            >>> uniform = gen.uniform(1, 10, size=10)
            >>> norm = gen.normal(0, 3, size=20)

        There is also a shorthand syntax that may be convenient.
        The property `np` is equivalent to this method, so you can also write::

            >>> uniform = rng.np.uniform(1, 10, size=10)
            >>> norm = rng.np.normal(0, 3, size=20)
        """
        return np.random.Generator(GalSimBitGenerator(self))

    def duplicate(self):
        """Create a duplicate of the current `BaseDeviate` object.

        The subsequent series from each copy of the `BaseDeviate` will produce identical values::

            >>> u = galsim.UniformDeviate(31415926)
            >>> u()
            0.17100770119577646
            >>> u2 = u.duplicate()
            >>> u()
            0.49095047544687986
            >>> u()
            0.10306670609861612
            >>> u2()
            0.49095047544687986
            >>> u2()
            0.10306670609861612
            >>> u2()
            0.13129289541393518
            >>> u()
            0.13129289541393518
        """
        ret = BaseDeviate.__new__(self.__class__)
        ret.__dict__.update(self.__dict__)
        ret._rng = self._rng.duplicate()
        return ret

    def __copy__(self):
        return self.duplicate()

    def __getstate__(self):
        d = self.__dict__.copy()
        d['rng_str'] = self.serialize()
        d.pop('_rng')
        return d

    def __setstate__(self, d):
        self.__dict__ = d
        rng = _galsim.BaseDeviateImpl(d['rng_str'])
        self._rng = self._rng_type(rng, *self._rng_args)

    def clearCache(self):
        """Clear the internal cache of the `BaseDeviate`, if any.  This is currently only relevant
        for `GaussianDeviate`, since it generates two values at a time, saving the second one to
        use for the next output value.
        """
        self._rng.clearCache()

    def discard(self, n, suppress_warnings=False):
        """Discard n values from the current sequence of pseudo-random numbers.

        This is typically used to keep two random number generators in sync when one of them
        is used to generate some random values.  The other can quickly discard the same number
        of random values to stay in sync with the first one.

        Parameters:
            n:                  The number of raw random numbers to discard.
            suppress_warnings:  Whether to suppress warnings related to detecting when this
                                action is not likely to reliably keep two random number
                                generators in sync. [default: False]
        """
        if not self.has_reliable_discard and not suppress_warnings:
            galsim_warn(self.__class__.__name__ +
                        " does not use a consistent number of randoms per generated value, " +
                        "so discard cannot be guaranteed to keep two random deviates in sync.")
        if n%2 == 1 and self.generates_in_pairs and not suppress_warnings:
            galsim_warn(self.__class__.__name__ +
                        " uses two randoms per pair of generated values, so discarding " +
                        "an odd number of randoms probably doesn't make sense.")
        self._rng.discard(int(n))

    @property
    def has_reliable_discard(self):
        """Indicates whether the generator always uses 1 rng per value.

        If it does, then `discard` can reliably be used to keep two generators in sync when one
        of them generated some values and the other didn't.

        This is False for PoissonDeviate, Chi2Deviate, and GammaDeviate.

        See also `BaseDeviate.generates_in_pairs`.
        """
        return True

    @property
    def generates_in_pairs(self):
        """Indicates whether the generator uses 2 rngs values per 2 returned values.

        GaussianDeviate has a slight wrinkle to the `BaseDeviate.has_reliable_discard` story.
        It always uses two rng values to generate two Gaussian deviates.  So if the number of
        generated values is even, then discard can keep things in sync.  However, if an odd
        number of values are generated, then you to generate one more value (and discard it)
        to keep things synced up.

        This is only True for GaussianDeviate.
        """
        return False

    def raw(self):
        """Generate the next pseudo-random number and rather than return the appropriate kind
        of random deviate for this class, just return the raw integer value that would have been
        used to generate this value.
        """
        return int(self._rng.raw())

    def generate(self, array):
        """Generate many pseudo-random values, filling in the values of a numpy array.
        """
        array_1d = np.ascontiguousarray(array.ravel(),dtype=float)
        #assert(array_1d.strides[0] == array_1d.itemsize)
        _a = array_1d.__array_interface__['data'][0]
        self._rng.generate(len(array_1d), _a)
        if array_1d.data != array.data:
            # array_1d is not a view into the original array.  Need to copy back.
            np.copyto(array, array_1d.reshape(array.shape), casting='unsafe')

    def add_generate(self, array):
        """Generate many pseudo-random values, adding them to the values of a numpy array.
        """
        array_1d = np.ascontiguousarray(array.ravel(),dtype=float)
        #assert(array_1d.strides[0] == array_1d.itemsize)
        _a = array_1d.__array_interface__['data'][0]
        self._rng.add_generate(len(array_1d), _a)
        if array_1d.data != array.data:
            # array_1d is not a view into the original array.  Need to copy back.
            np.copyto(array, array_1d.reshape(array.shape), casting='unsafe')

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, self.__class__) and
                 self._rng_type == other._rng_type and
                 self._rng_args == other._rng_args and
                 self.serialize() == other.serialize()))

    def __ne__(self, other):
        return not self.__eq__(other)

    __hash__ = None

    def serialize(self):
        return str(self._rng.serialize())

    def _seed_repr(self):
        s = self.serialize().split(' ')
        return " ".join(s[:3])+" ... "+" ".join(s[-3:])

    def __repr__(self):
        return "galsim.BaseDeviate(%r)"%self._seed_repr()

    def __str__(self):
        return "galsim.BaseDeviate(%r)"%self._seed_repr()

class UniformDeviate(BaseDeviate):
    """Pseudo-random number generator with uniform distribution in interval [0.,1.).

    Successive calls to ``u()`` generate pseudo-random values distributed uniformly in the interval
    [0., 1.)::

        >>> u = galsim.UniformDeviate(31415926)
        >>> u()
        0.17100770119577646
        >>> u()
        0.49095047544687986

    Parameters:
        seed:       Something that can seed a `BaseDeviate`: an integer seed or another
                    `BaseDeviate`.  Using 0 means to generate a seed from the system.
                    [default: None]

    """
    def __init__(self, seed=None):
        self._rng_type = _galsim.UniformDeviateImpl
        self._rng_args = ()
        self.reset(seed)

    def __call__(self):
        """Draw a new random number from the distribution.

        Returns a uniform deviate between 0 and 1.
        """
        return self._rng.generate1()

    def __repr__(self):
        return 'galsim.UniformDeviate(seed=%r)'%(self._seed_repr())
    def __str__(self):
        return 'galsim.UniformDeviate()'


class GaussianDeviate(BaseDeviate):
    """Pseudo-random number generator with Gaussian distribution.

    See http://en.wikipedia.org/wiki/Gaussian_distribution for further details.

    Successive calls to ``g()`` generate pseudo-random values distributed according to a Gaussian
    distribution with the provided ``mean``, ``sigma``::

        >>> g = galsim.GaussianDeviate(31415926)
        >>> g()
        0.5533754000847082
        >>> g()
        1.0218588970190354

    Parameters:
        seed:       Something that can seed a `BaseDeviate`: an integer seed or another
                    `BaseDeviate`.  Using 0 means to generate a seed from the system.
                    [default: None]
        mean:       Mean of Gaussian distribution. [default: 0.]
        sigma:      Sigma of Gaussian distribution. [default: 1.; Must be > 0]
    """
    def __init__(self, seed=None, mean=0., sigma=1.):
        if sigma < 0.:
            raise GalSimRangeError("GaussianDeviate sigma must be > 0.", sigma, 0.)
        self._rng_type = _galsim.GaussianDeviateImpl
        self._rng_args = (float(mean), float(sigma))
        self.reset(seed)

    @property
    def mean(self):
        """The mean of the Gaussian distribution.
        """
        return self._rng_args[0]

    @property
    def sigma(self):
        """The sigma of the Gaussian distribution.
        """
        return self._rng_args[1]

    @property
    def generates_in_pairs(self):
        return True

    def __call__(self):
        """Draw a new random number from the distribution.

        Returns a Gaussian deviate with the given mean and sigma.
        """
        return self._rng.generate1()

    def generate_from_variance(self, array):
        """Generate many Gaussian deviate values using the existing array values as the
        variance for each.
        """
        array_1d = np.ascontiguousarray(array.ravel(), dtype=float)
        #assert(array_1d.strides[0] == array_1d.itemsize)
        _a = array_1d.__array_interface__['data'][0]
        self._rng.generate_from_variance(len(array_1d), _a)
        if array_1d.data != array.data:
            # array_1d is not a view into the original array.  Need to copy back.
            np.copyto(array, array_1d.reshape(array.shape), casting='unsafe')

    def __repr__(self):
        return 'galsim.GaussianDeviate(seed=%r, mean=%r, sigma=%r)'%(
                self._seed_repr(), self.mean, self.sigma)
    def __str__(self):
        return 'galsim.GaussianDeviate(mean=%r, sigma=%r)'%(self.mean, self.sigma)


class BinomialDeviate(BaseDeviate):
    """Pseudo-random Binomial deviate for ``N`` trials each of probability ``p``.

    ``N`` is number of 'coin flips,' ``p`` is probability of 'heads,' and each call returns an
    integer value where 0 <= value <= N gives the number of heads.  See
    http://en.wikipedia.org/wiki/Binomial_distribution for more information.

    Successive calls to ``b()`` generate pseudo-random integer values distributed according to a
    binomial distribution with the provided ``N``, ``p``::

        >>> b = galsim.BinomialDeviate(31415926, N=10, p=0.3)
        >>> b()
        2
        >>> b()
        3

    Parameters:
        seed:       Something that can seed a `BaseDeviate`: an integer seed or another
                    `BaseDeviate`.  Using 0 means to generate a seed from the system.
                    [default: None]
        N:          The number of 'coin flips' per trial. [default: 1; Must be > 0]
        p:          The probability of success per coin flip. [default: 0.5; Must be > 0]
    """
    def __init__(self, seed=None, N=1, p=0.5):
        self._rng_type = _galsim.BinomialDeviateImpl
        self._rng_args = (int(N), float(p))
        self.reset(seed)

    @property
    def n(self):
        """The number of 'coin flips'.
        """
        return self._rng_args[0]

    @property
    def p(self):
        """The probability of success per 'coin flip'.
        """
        return self._rng_args[1]

    def __call__(self):
        """Draw a new random number from the distribution.

        Returns a Binomial deviate with the given n and p.
        """
        return self._rng.generate1()

    def __repr__(self):
        return 'galsim.BinomialDeviate(seed=%r, N=%r, p=%r)'%(self._seed_repr(), self.n, self.p)
    def __str__(self):
        return 'galsim.BinomialDeviate(N=%r, p=%r)'%(self.n, self.p)


class PoissonDeviate(BaseDeviate):
    """Pseudo-random Poisson deviate with specified ``mean``.

    The input ``mean`` sets the mean and variance of the Poisson deviate.  An integer deviate with
    this distribution is returned after each call.
    See http://en.wikipedia.org/wiki/Poisson_distribution for more details.

    Successive calls to ``p()`` generate pseudo-random integer values distributed according to a
    Poisson distribution with the specified ``mean``::

        >>> p = galsim.PoissonDeviate(31415926, mean=100)
        >>> p()
        94
        >>> p()
        106

    Parameters:
        seed:       Something that can seed a `BaseDeviate`: an integer seed or another
                    `BaseDeviate`.  Using 0 means to generate a seed from the system.
                    [default: None]
        mean:       Mean of the distribution. [default: 1; Must be > 0]
    """
    def __init__(self, seed=None, mean=1.):
        if mean < 0:
            raise GalSimValueError("PoissonDeviate is only defined for mean >= 0.", mean)
        self._rng_type = _galsim.PoissonDeviateImpl
        self._rng_args = (float(mean),)
        self.reset(seed)

    @property
    def mean(self):
        """The mean of the distribution.
        """
        return self._rng_args[0]

    @property
    def has_reliable_discard(self):
        return False

    def __call__(self):
        """Draw a new random number from the distribution.

        Returns a Poisson deviate with the given mean.
        """
        return self._rng.generate1()

    def generate_from_expectation(self, array):
        """Generate many Poisson deviate values using the existing array values as the
        expectation value (aka mean) for each.
        """
        if np.any(array < 0):
            raise GalSimValueError("Expectation array may not have values < 0.", array)
        array_1d = np.ascontiguousarray(array.ravel(), dtype=float)
        #assert(array_1d.strides[0] == array_1d.itemsize)
        _a = array_1d.__array_interface__['data'][0]
        self._rng.generate_from_expectation(len(array_1d), _a)
        if array_1d.data != array.data:
            # array_1d is not a view into the original array.  Need to copy back.
            np.copyto(array, array_1d.reshape(array.shape), casting='unsafe')

    def __repr__(self):
        return 'galsim.PoissonDeviate(seed=%r, mean=%r)'%(self._seed_repr(), self.mean)
    def __str__(self):
        return 'galsim.PoissonDeviate(mean=%r)'%(self.mean)


class WeibullDeviate(BaseDeviate):
    """Pseudo-random Weibull-distributed deviate for shape parameter ``a`` and scale parameter ``b``.

    The Weibull distribution is related to a number of other probability distributions; in
    particular, it interpolates between the exponential distribution (a=1) and the Rayleigh
    distribution (a=2).
    See http://en.wikipedia.org/wiki/Weibull_distribution (a=k and b=lambda in the notation adopted
    in the Wikipedia article) for more details.  The Weibull distribution is real valued and
    produces deviates >= 0.

    Successive calls to ``w()`` generate pseudo-random values distributed according to a Weibull
    distribution with the specified shape and scale parameters ``a`` and ``b``::

        >>> w = galsim.WeibullDeviate(31415926, a=1.3, b=4)
        >>> w()
        1.1038481241018219
        >>> w()
        2.957052966368049

    Parameters:
        seed:       Something that can seed a `BaseDeviate`: an integer seed or another
                    `BaseDeviate`.  Using 0 means to generate a seed from the system.
                    [default: None]
        a:          Shape parameter of the distribution. [default: 1; Must be > 0]
        b:          Scale parameter of the distribution. [default: 1; Must be > 0]
    """
    def __init__(self, seed=None, a=1., b=1.):
        self._rng_type = _galsim.WeibullDeviateImpl
        self._rng_args = (float(a), float(b))
        self.reset(seed)

    @property
    def a(self):
        """The shape parameter, a.
        """
        return self._rng_args[0]

    @property
    def b(self):
        """The scale parameter, b.
        """
        return self._rng_args[1]

    def __call__(self):
        """Draw a new random number from the distribution.

        Returns a Weibull-distributed deviate with the given shape parameters a and b.
        """
        return self._rng.generate1()

    def __repr__(self):
        return 'galsim.WeibullDeviate(seed=%r, a=%r, b=%r)'%(self._seed_repr(), self.a, self.b)
    def __str__(self):
        return 'galsim.WeibullDeviate(a=%r, b=%r)'%(self.a, self.b)


class GammaDeviate(BaseDeviate):
    """A Gamma-distributed deviate with shape parameter ``k`` and scale parameter ``theta``.
    See http://en.wikipedia.org/wiki/Gamma_distribution.
    (Note: we use the k, theta notation. If you prefer alpha, beta, use k=alpha, theta=1/beta.)
    The Gamma distribution is a real valued distribution producing deviates >= 0.

    Successive calls to ``g()`` generate pseudo-random values distributed according to a gamma
    distribution with the specified shape and scale parameters ``k`` and ``theta``::

        >>> gam = galsim.GammaDeviate(31415926, k=1, theta=2)
        >>> gam()
        0.37508882726316
        >>> gam()
        1.3504199388358704

    Parameters:
        seed:       Something that can seed a `BaseDeviate`: an integer seed or another
                    `BaseDeviate`.  Using 0 means to generate a seed from the system.
                    [default: None]
        k:          Shape parameter of the distribution. [default: 1; Must be > 0]
        theta:      Scale parameter of the distribution. [default: 1; Must be > 0]
    """
    def __init__(self, seed=None, k=1., theta=1.):
        self._rng_type = _galsim.GammaDeviateImpl
        self._rng_args = (float(k), float(theta))
        self.reset(seed)

    @property
    def k(self):
        """The shape parameter, k.
        """
        return self._rng_args[0]

    @property
    def theta(self):
        """The scale parameter, theta.
        """
        return self._rng_args[1]

    @property
    def has_reliable_discard(self):
        return False

    def __call__(self):
        """Draw a new random number from the distribution.

        Returns a Gamma-distributed deviate with the given k and theta.
        """
        return self._rng.generate1()

    def __repr__(self):
        return 'galsim.GammaDeviate(seed=%r, k=%r, theta=%r)'%(
                self._seed_repr(), self.k, self.theta)
    def __str__(self):
        return 'galsim.GammaDeviate(k=%r, theta=%r)'%(self.k, self.theta)


class Chi2Deviate(BaseDeviate):
    """Pseudo-random Chi^2-distributed deviate for degrees-of-freedom parameter ``n``.

    See http://en.wikipedia.org/wiki/Chi-squared_distribution (note that k=n in the notation
    adopted in the Boost.Random routine called by this class).  The Chi^2 distribution is a
    real-valued distribution producing deviates >= 0.

    Successive calls to ``chi2()`` generate pseudo-random values distributed according to a
    chi-square distribution with the specified degrees of freedom, ``n``::

        >>> chi2 = galsim.Chi2Deviate(31415926, n=7)
        >>> chi2()
        7.9182211987712385
        >>> chi2()
        6.644121724269535

    Parameters:
        seed:       Something that can seed a `BaseDeviate`: an integer seed or another
                    `BaseDeviate`.  Using 0 means to generate a seed from the system.
                    [default: None]
        n:          Number of degrees of freedom for the output distribution. [default: 1;
                    Must be > 0]
    """
    def __init__(self, seed=None, n=1.):
        self._rng_type = _galsim.Chi2DeviateImpl
        self._rng_args = (float(n),)
        self.reset(seed)

    @property
    def n(self):
        """The number of degrees of freedom.
        """
        return self._rng_args[0]

    @property
    def has_reliable_discard(self):
        return False

    def __call__(self):
        """Draw a new random number from the distribution.

        Returns a Chi2-distributed deviate with the given number of degrees of freedom.
        """
        return self._rng.generate1()

    def __repr__(self):
        return 'galsim.Chi2Deviate(seed=%r, n=%r)'%(self._seed_repr(), self.n)
    def __str__(self):
        return 'galsim.Chi2Deviate(n=%r)'%(self.n)


class DistDeviate(BaseDeviate):
    """A class to draw random numbers from a user-defined probability distribution.

    DistDeviate is a `BaseDeviate` class that can be used to draw from an arbitrary probability
    distribution.  The probability distribution passed to DistDeviate can be given one of three
    ways: as the name of a file containing a 2d ASCII array of x and P(x), as a `LookupTable`
    mapping x to P(x), or as a callable function.

    Once given a probability, DistDeviate creates a table of the cumulative probability and draws
    from it using a `UniformDeviate`.  The precision of its outputs can be controlled with the
    keyword ``npoints``, which sets the number of points DistDeviate creates for its internal table
    of CDF(x).  To prevent errors due to non-monotonicity, the interpolant for this internal table
    is always linear.

    Two keywords, ``x_min`` and ``x_max``, define the support of the function.  They must be passed
    if a callable function is given to DistDeviate, unless the function is a `LookupTable`, which
    has its own defined endpoints.  If a filename or `LookupTable` is passed to DistDeviate, the
    use of ``x_min`` or ``x_max`` will result in an error.

    If given a table in a file, DistDeviate will construct an interpolated `LookupTable` to obtain
    more finely gridded probabilities for generating the cumulative probability table.  The default
    ``interpolant`` is linear, but any interpolant understood by `LookupTable` may be used.  We
    caution against the use of splines because they can cause non-monotonic behavior.  Passing the
    ``interpolant`` keyword next to anything but a table in a file will result in an error.

    **Examples**:

    Some sample initialization calls::

        >>> d = galsim.DistDeviate(function=f, x_min=x_min, x_max=x_max)

    Initializes d to be a DistDeviate instance with a distribution given by the callable function
    ``f(x)`` from ``x=x_min`` to ``x=x_max`` and seeds the PRNG using current time::

        >>> d = galsim.DistDeviate(1062533, function=file_name, interpolant='floor')

    Initializes d to be a DistDeviate instance with a distribution given by the data in file
    ``file_name``, which must be a 2-column ASCII table, and seeds the PRNG using the integer
    seed 1062533. It generates probabilities from ``file_name`` using the interpolant 'floor'::

        >>> d = galsim.DistDeviate(rng, function=galsim.LookupTable(x,p))

    Initializes d to be a DistDeviate instance with a distribution given by P(x), defined as two
    arrays ``x`` and ``p`` which are used to make a callable `LookupTable`, and links the
    DistDeviate PRNG to the already-existing random number generator ``rng``.

    Successive calls to ``d()`` generate pseudo-random values with the given probability
    distribution::

        >>> d = galsim.DistDeviate(31415926, function=lambda x: 1-abs(x), x_min=-1, x_max=1)
        >>> d()
        -0.4151921102709466
        >>> d()
        -0.00909781188974034

    Parameters:
        seed:           Something that can seed a `BaseDeviate`: an integer seed or another
                        `BaseDeviate`.  Using 0 means to generate a seed from the system.
                        [default: None]
        function:       A callable function giving a probability distribution or the name of a
                        file containing a probability distribution as a 2-column ASCII table.
                        [required]
        x_min:          The minimum desired return value (required for non-`LookupTable`
                        callable functions; will raise an error if not passed in that case, or if
                        passed in any other case) [default: None]
        x_max:          The maximum desired return value (required for non-`LookupTable`
                        callable functions; will raise an error if not passed in that case, or if
                        passed in any other case) [default: None]
        interpolant:    Type of interpolation used for interpolating a file (causes an error if
                        passed alongside a callable function).  Options are given in the
                        documentation for `LookupTable`. [default: 'linear']
        npoints:        Number of points DistDeviate should create for its internal interpolation
                        tables. [default: 256, unless the function is a non-log `LookupTable`, in
                        which case it uses the table's x values]
        clip_neg:       Clip any negative input values to zero. [default: False; an error will
                        be raised if any negative probabilities are found.]
    """
    def __init__(self, seed=None, function=None, x_min=None,
                 x_max=None, interpolant=None, npoints=None, clip_neg=False):

        # Set up the PRNG
        self._rng_type = _galsim.UniformDeviateImpl
        self._rng_args = ()
        self.reset(seed)

        # Basic input checking and setups
        if function is None:
            raise TypeError('You must pass a function to DistDeviate!')

        self._interpolant = interpolant
        self._npoints = npoints
        self._xmin = x_min
        self._xmax = x_max

        # Figure out if a string is a filename or something we should be using in an eval call
        if isinstance(function, str):
            self._function = function # Save the inputs to be used in repr
            if os.path.isfile(function):
                if interpolant is None:
                    interpolant='linear'
                if x_min or x_max:
                    raise GalSimIncompatibleValuesError(
                        "Cannot pass x_min or x_max with a filename argument",
                        function=function, x_min=x_min, x_max=x_max)
                function = LookupTable.from_file(function, interpolant=interpolant)
                x_min = function.x_min
                x_max = function.x_max
            else:
                try:
                    function = math_eval('lambda x : ' + function)
                    if x_min is not None: # is not None in case x_min=0.
                        function(x_min)
                    else:
                        # Somebody would be silly to pass a string for evaluation without x_min,
                        # but we'd like to throw reasonable errors in that case anyway
                        function(0.6) # A value unlikely to be a singular point of a function
                except Exception as e:
                    raise GalSimValueError(
                        "String function must either be a valid filename or something that "
                        "can eval to a function of x.\n"
                        "Caught error: {0}".format(e), self._function)
        else:
            # Check that the function is actually a function
            if not hasattr(function, '__call__'):
                raise TypeError('function must be a callable function or a string')
            if interpolant:
                raise GalSimIncompatibleValuesError(
                    "Cannot provide an interpolant with a callable function argument",
                    interpolant=interpolant, function=function)
            if isinstance(function, LookupTable):
                if (x_min not in (None, function.x_min)) or (x_max not in (None, function.x_max)):
                    raise GalSimIncompatibleValuesError(
                        "Cannot provide x_min or x_max with a LookupTable function",
                        function=function, x_min=x_min, x_max=x_max)
                x_min = function.x_min
                x_max = function.x_max
            else:
                if x_min is None or x_max is None:
                    raise GalSimIncompatibleValuesError(
                        "Must provide x_min and x_max when function argument is a regular "
                        "python callable function",
                        function=function, x_min=x_min, x_max=x_max)

            self._function = function # Save the inputs to be used in repr

        # Compute the probability distribution function, pdf(x)
        if (npoints is None and isinstance(function, LookupTable) and
            not function.x_log and not function.f_log):
            xarray = np.array(function.x, dtype=float)
            pdf = np.array(function.f, dtype=float)
            # Set up pdf, so cumsum basically does a cumulative trapz integral
            # On Python 3.4, doing pdf[1:] += pdf[:-1] the last value gets messed up.
            # Writing it this way works.  (Maybe slightly slower though, so if we stop
            # supporting python 3.4, consider switching to the += version.)
            pdf[1:] = pdf[1:] + pdf[:-1]
            pdf[1:] *= np.diff(xarray)
            pdf[0] = 0.
        else:
            if npoints is None: npoints = 256
            xarray = x_min+(1.*x_max-x_min)/(npoints-1)*np.array(range(npoints),float)
            # Integrate over the range of x in case the function is doing something weird here.
            pdf = [0.] + [integ.int1d(function, xarray[i], xarray[i+1])
                          for i in range(npoints - 1)]
            pdf = np.array(pdf)

        # Check that the probability is nonnegative
        if clip_neg:
            # Write it this way so nan -> 0 as well as negative values.
            w = np.where(~(pdf >= 0))
            pdf[w] = 0.
        elif not np.all(pdf >= 0.):
            raise GalSimValueError('Negative probability found in DistDeviate.',function)

        # Compute the cumulative distribution function = int(pdf(x),x)
        cdf = np.cumsum(pdf)

        # Quietly renormalize the probability if it wasn't already normalized
        totalprobability = cdf[-1]
        cdf /= totalprobability

        self._inverse_cdf = LookupTable(cdf, xarray, interpolant='linear')
        self.x_min = x_min
        self.x_max = x_max

    def val(self, p):
        r"""
        Return the value :math:`x` of the input function to `DistDeviate` such that ``p`` =
        :math:`F(x)`, where :math:`F` is the cumulattive probability distribution function:

        .. math::

            F(x) = \int_{-\infty}^x \mathrm{pdf}(t) dt

        This function is typically called by `__call__`, which generates a random p
        between 0 and 1 and calls ``self.val(p)``.

        Parameters:
            p:      The desired cumulative probabilty p.

        Returns:
            the corresponding x such that :math:`p = F(x)`.
        """
        if p<0 or p>1:
            raise GalSimRangeError('Invalid cumulative probability for DistDeviate', p, 0., 1.)
        return self._inverse_cdf(p)

    def __call__(self):
        """Draw a new random number from the distribution.
        """
        return self._inverse_cdf(self._rng.generate1())

    def generate(self, array):
        """Generate many pseudo-random values, filling in the values of a numpy array.
        """
        p = np.empty_like(array)
        BaseDeviate.generate(self, p)  # Fill with unform deviate values
        np.copyto(array, self._inverse_cdf(p)) # Convert from p -> x

    def add_generate(self, array):
        """Generate many pseudo-random values, adding them to the values of a numpy array.
        """
        p = np.empty_like(array)
        BaseDeviate.generate(self, p)
        array += self._inverse_cdf(p)

    def __repr__(self):
        return ('galsim.DistDeviate(seed=%r, function=%r, x_min=%r, x_max=%r, interpolant=%r, '
                'npoints=%r)')%(self._seed_repr(), self._function, self._xmin, self._xmax,
                                self._interpolant, self._npoints)
    def __str__(self):
        return 'galsim.DistDeviate(function="%s", x_min=%s, x_max=%s, interpolant=%s, npoints=%s)'%(
                self._function, self._xmin, self._xmax, self._interpolant, self._npoints)

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, DistDeviate) and
                 self.serialize() == other.serialize() and
                 self._function == other._function and
                 self._xmin == other._xmin and
                 self._xmax == other._xmax and
                 self._interpolant == other._interpolant and
                 self._npoints == other._npoints))


class GalSimBitGenerator(np.random.BitGenerator):
    """A numpy.random.BitGenerator that uses the GalSim C++-layer random number generator
    for the random bit generation.

    Parameters:
        rng:    The galsim.BaseDeviate object to use for the underlying bit generation.
    """
    def __init__(self, rng):
        super().__init__(0)
        self.rng = rng
        self.rng._rng.setup_bitgen(self.capsule)

def permute(rng, *args):
    """Randomly permute one or more lists.

    If more than one list is given, then all lists will have the same random permutation
    applied to it.

    Parameters:
        rng:    The random number generator to use. (This will be converted to a `UniformDeviate`.)
        args:   Any number of lists to be permuted.
    """
    ud = UniformDeviate(rng)
    if len(args) == 0:
        raise TypeError("permute called with no lists to permute")

    # We use an algorithm called the Knuth shuffle, which is based on the Fisher-Yates shuffle.
    # See http://en.wikipedia.org/wiki/Fisher-Yates_shuffle for more information.
    n = len(args[0])
    for i in range(n-1,1,-1):
        j = int((i+1) * ud())
        if j == i+1: j = i  # I'm not sure if this is possible, but just in case...
        for lst in args:
            lst[i], lst[j] = lst[j], lst[i]



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import os
import warnings
import numpy as np

import galsim
from galsim_test_helpers import *


def check_dep(f, *args, **kwargs):
    """Check that some function raises a GalSimDeprecationWarning as a warning, but not an error.
    """
    # Check that f() raises a warning, but not an error.
    with assert_warns(galsim.GalSimDeprecationWarning):
        res = f(*args, **kwargs)
    return res


@timer
def test_gsparams():
    check_dep(galsim.GSParams, allowed_flux_variation=0.90)
    check_dep(galsim.GSParams, range_division_for_extrema=50)
    check_dep(galsim.GSParams, small_fraction_of_flux=1.e-6)


@timer
def test_phase_psf():
    atm = galsim.Atmosphere(screen_size=10.0, altitude=0, r0_500=0.15, suppress_warning=True)
    psf = atm.makePSF(exptime=0.02, time_step=0.01, diam=1.1, lam=1000.0)
    check_dep(galsim.PhaseScreenPSF.__getattribute__, psf, "img")
    check_dep(galsim.PhaseScreenPSF.__getattribute__, psf, "finalized")

@timer
def test_interpolant():
    d = check_dep(galsim.Delta, tol=1.e-2)
    assert d.gsparams.kvalue_accuracy == 1.e-2
    assert check_dep(getattr, d, 'tol') == d.gsparams.kvalue_accuracy
    n = check_dep(galsim.Nearest, tol=1.e-2)
    assert n.gsparams.kvalue_accuracy == 1.e-2
    assert check_dep(getattr, n, 'tol') == n.gsparams.kvalue_accuracy
    s = check_dep(galsim.SincInterpolant, tol=1.e-2)
    assert s.gsparams.kvalue_accuracy == 1.e-2
    assert check_dep(getattr, s, 'tol') == s.gsparams.kvalue_accuracy
    l = check_dep(galsim.Linear, tol=1.e-2)
    assert l.gsparams.kvalue_accuracy == 1.e-2
    assert check_dep(getattr, l, 'tol') == l.gsparams.kvalue_accuracy
    c = check_dep(galsim.Cubic, tol=1.e-2)
    assert c.gsparams.kvalue_accuracy == 1.e-2
    assert check_dep(getattr, c, 'tol') == c.gsparams.kvalue_accuracy
    q = check_dep(galsim.Quintic, tol=1.e-2)
    assert q.gsparams.kvalue_accuracy == 1.e-2
    assert check_dep(getattr, q, 'tol') == q.gsparams.kvalue_accuracy
    l3 = check_dep(galsim.Lanczos, 3, tol=1.e-2)
    assert l3.gsparams.kvalue_accuracy == 1.e-2
    assert check_dep(getattr, l3, 'tol') == l3.gsparams.kvalue_accuracy
    ldc = check_dep(galsim.Lanczos, 3, False, tol=1.e-2)
    assert ldc.gsparams.kvalue_accuracy == 1.e-2
    assert check_dep(getattr, ldc, 'tol') == ldc.gsparams.kvalue_accuracy
    l8 = check_dep(galsim.Lanczos, 8, tol=1.e-2)
    assert l8.gsparams.kvalue_accuracy == 1.e-2
    assert check_dep(getattr, l8, 'tol') == l8.gsparams.kvalue_accuracy
    l11 = check_dep(galsim.Interpolant.from_name, 'lanczos11', tol=1.e-2)
    assert l11.gsparams.kvalue_accuracy == 1.e-2
    assert check_dep(getattr, l11, 'tol') == l11.gsparams.kvalue_accuracy

@timer
def test_noise():
    real_gal_dir = os.path.join('..','examples','data')
    real_gal_cat = 'real_galaxy_catalog_23.5_example.fits'
    real_cat = galsim.RealGalaxyCatalog(
        dir=real_gal_dir, file_name=real_gal_cat, preload=True)

    test_seed=987654
    test_index = 17
    cf_1 = real_cat.getNoise(test_index, rng=galsim.BaseDeviate(test_seed))
    im_2, pix_scale_2, var_2 = real_cat.getNoiseProperties(test_index)
    # Check the variance:
    var_1 = cf_1.getVariance()
    assert var_1==var_2,'Inconsistent noise variance from getNoise and getNoiseProperties'
    # Check the image:
    ii = galsim.InterpolatedImage(im_2, normalization='sb', calculate_stepk=False,
                                  calculate_maxk=False, x_interpolant='linear')
    cf_2 = check_dep(galsim.correlatednoise._BaseCorrelatedNoise,
                     galsim.BaseDeviate(test_seed), ii, im_2.wcs)
    cf_2 = cf_2.withVariance(var_2)
    assert cf_1==cf_2,'Inconsistent noise properties from getNoise and getNoiseProperties'

@timer
def test_randwalk_defaults():
    """
    Create a random walk galaxy and test that the getters work for
    default inputs
    """

    # try constructing with mostly defaults
    npoints=100
    hlr = 8.0
    rng = galsim.BaseDeviate(1234)
    rw=check_dep(galsim.RandomWalk, npoints, half_light_radius=hlr, rng=rng)

    assert rw.npoints==npoints,"expected npoints==%d, got %d" % (npoints, rw.npoints)
    assert rw.input_half_light_radius==hlr,\
        "expected hlr==%g, got %g" % (hlr, rw.input_half_light_radius)

    nobj=len(rw.points)
    assert nobj == npoints,"expected %d objects, got %d" % (npoints, nobj)

    pts=rw.points
    assert pts.shape == (npoints,2),"expected (%d,2) shape for points, got %s" % (npoints, pts.shape)
    np.testing.assert_almost_equal(rw.centroid.x, np.mean(pts[:,0]))
    np.testing.assert_almost_equal(rw.centroid.y, np.mean(pts[:,1]))

    gsp = galsim.GSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)
    rng2 = galsim.BaseDeviate(1234)
    rw2 = check_dep(galsim.RandomWalk, npoints, half_light_radius=hlr, rng=rng2, gsparams=gsp)
    assert rw2 != rw
    assert rw2 == rw.withGSParams(gsp)

    # Check that they produce identical images.
    psf = galsim.Gaussian(sigma=0.8)
    conv1 = galsim.Convolve(rw.withGSParams(gsp), psf)
    conv2 = galsim.Convolve(rw2, psf)
    im1 = conv1.drawImage()
    im2 = conv2.drawImage()
    assert im1 == im2

    # Check that image is not sensitive to use of rng by other objects.
    rng3 = galsim.BaseDeviate(1234)
    rw3=check_dep(galsim.RandomWalk, npoints, half_light_radius=hlr, rng=rng3)
    rng3.discard(523)
    conv1 = galsim.Convolve(rw, psf)
    conv3 = galsim.Convolve(rw3, psf)
    im1 = conv1.drawImage()
    im3 = conv2.drawImage()
    assert im1 == im3

    # Run some basic tests of correctness
    check_basic(conv1, "RandomWalk")
    im = galsim.ImageD(64,64, scale=0.5)
    do_shoot(conv1, im, "RandomWalk")
    do_kvalue(conv1, im, "RandomWalk")
    check_pickle(rw)
    check_pickle(conv1)
    check_pickle(conv1, lambda x: x.drawImage(scale=1))


@timer
def test_randwalk_repr():
    """
    test the repr and str work, and that a new object can be created
    using eval
    """

    npoints=100
    hlr = 8.0
    flux=1
    rw1=check_dep(galsim.RandomWalk,
        npoints,
        half_light_radius=hlr,
        flux=flux,
    )
    rw2=check_dep(galsim.RandomWalk,
        npoints,
        profile=galsim.Exponential(half_light_radius=hlr, flux=flux),
    )

    for rw in (rw1, rw2):


        # just make sure str() works, don't require eval to give
        # a consistent object back
        st=str(rw)

        # require eval(repr(rw)) to give a consistent object back

        new_rw = eval(repr(rw))

        assert new_rw.npoints == rw.npoints,\
            "expected npoints=%d got %d" % (rw.npoints,new_rw.npoints)

        mess="expected input_half_light_radius=%.16g got %.16g"
        assert new_rw.input_half_light_radius == rw.input_half_light_radius,\
            mess % (rw.input_half_light_radius,new_rw.input_half_light_radius)
        assert new_rw.flux == rw.flux,\
            "expected flux=%.16g got %.16g" % (rw.flux,new_rw.flux)

@timer
def test_randwalk_config():
    """
    test we get the same object using a configuration and the
    explicit constructor
    """

    hlr=2.0
    flux=np.pi
    gal_config1 = {
        'type':'RandomWalk',
        'npoints':100,
        'half_light_radius':hlr,
        'flux':flux,
    }
    gal_config2 = {
        'type':'RandomWalk',
        'npoints':150,
        'profile': {
            'type': 'Exponential',
            'half_light_radius': hlr,
            'flux': flux,
        }
    }

    for gal_config in (gal_config1, gal_config2):
        config={
            'gal':gal_config,
            'rng':galsim.BaseDeviate(31415),
        }

        rwc = check_dep(galsim.config.BuildGSObject, config, 'gal')[0]
        print(repr(rwc._profile))

        rw = check_dep(galsim.RandomWalk,
            gal_config['npoints'],
            half_light_radius=hlr,
            flux=flux,
        )

        assert rw.npoints==rwc.npoints,\
            "expected npoints==%d, got %d" % (rw.npoints, rwc.npoints)

        assert rw.input_half_light_radius==rwc.input_half_light_radius,\
            "expected hlr==%g, got %g" % (rw.input_half_light_radius, rw.input_half_light_radius)

        nobj=len(rw.points)
        nobjc=len(rwc.points)
        assert nobj==nobjc,"expected %d objects, got %d" % (nobj,nobjc)

        pts=rw.points
        ptsc=rwc.points
        assert (pts.shape == ptsc.shape),\
                "expected %s shape for points, got %s" % (pts.shape,ptsc.shape)

@timer
def test_withOrigin():
    from test_wcs import Cubic

    # First EuclideantWCS types:

    wcs_list = [ galsim.OffsetWCS(0.3, galsim.PositionD(1,1), galsim.PositionD(10,23)),
                 galsim.OffsetShearWCS(0.23, galsim.Shear(g1=0.1,g2=0.3), galsim.PositionD(12,43)),
                 galsim.AffineTransform(0.01,0.26,-0.26,0.02, galsim.PositionD(12,43)),
                 galsim.UVFunction(ufunc = lambda x,y: 0.2*x, vfunc = lambda x,y: 0.2*y),
                 galsim.UVFunction(ufunc = lambda x,y: 0.2*x, vfunc = lambda x,y: 0.2*y,
                                   xfunc = lambda u,v: u / scale, yfunc = lambda u,v: v / scale),
                 galsim.UVFunction(ufunc='0.2*x + 0.03*y', vfunc='0.01*x + 0.2*y'),
               ]

    color = 0.3
    for wcs in wcs_list:
        # Original version of the shiftOrigin tests in do_nonlocal_wcs using deprecated name.
        new_origin = galsim.PositionI(123,321)
        wcs3 = check_dep(wcs.withOrigin, new_origin)
        assert wcs != wcs3, name+' is not != wcs.withOrigin(pos)'
        wcs4 = wcs.local(wcs.origin, color=color)
        assert wcs != wcs4, name+' is not != wcs.local()'
        assert wcs4 != wcs, name+' is not != wcs.local() (reverse)'
        world_origin = wcs.toWorld(wcs.origin, color=color)
        if wcs.isUniform():
            if wcs.world_origin == galsim.PositionD(0,0):
                wcs2 = wcs.local(wcs.origin, color=color).withOrigin(wcs.origin)
                assert wcs == wcs2, name+' is not equal after wcs.local().withOrigin(origin)'
            wcs2 = wcs.local(wcs.origin, color=color).withOrigin(wcs.origin, wcs.world_origin)
            assert wcs == wcs2, name+' not equal after wcs.local().withOrigin(origin,world_origin)'
        world_pos1 = wcs.toWorld(galsim.PositionD(0,0), color=color)
        wcs3 = check_dep(wcs.withOrigin, new_origin)
        world_pos2 = wcs3.toWorld(new_origin, color=color)
        np.testing.assert_almost_equal(
                world_pos2.x, world_pos1.x, 7,
                'withOrigin(new_origin) returned wrong world position')
        np.testing.assert_almost_equal(
                world_pos2.y, world_pos1.y, 7,
                'withOrigin(new_origin) returned wrong world position')
        new_world_origin = galsim.PositionD(5352.7, 9234.3)
        wcs5 = check_dep(wcs.withOrigin, new_origin, new_world_origin, color=color)
        world_pos3 = wcs5.toWorld(new_origin, color=color)
        np.testing.assert_almost_equal(
                world_pos3.x, new_world_origin.x, 7,
                'withOrigin(new_origin, new_world_origin) returned wrong position')
        np.testing.assert_almost_equal(
                world_pos3.y, new_world_origin.y, 7,
                'withOrigin(new_origin, new_world_origin) returned wrong position')

    # Now some CelestialWCS types
    cubic_u = Cubic(2.9e-5, 2000., 'u')
    cubic_v = Cubic(-3.7e-5, 2000., 'v')
    center = galsim.CelestialCoord(23 * galsim.degrees, -13 * galsim.degrees)
    radec = lambda x,y: center.deproject_rad(cubic_u(x,y)*0.2, cubic_v(x,y)*0.2,
                                             projection='lambert')
    wcs_list = [ galsim.RaDecFunction(radec),
                 galsim.AstropyWCS('1904-66_TAN.fits', dir='fits_files'),
                 galsim.GSFitsWCS('tpv.fits', dir='fits_files'),
                 galsim.FitsWCS('sipsample.fits', dir='fits_files'),
               ]

    for wcs in wcs_list:
        # Original version of the shiftOrigin tests in do_celestial_wcs using deprecated name.
        new_origin = galsim.PositionI(123,321)
        wcs3 = wcs.shiftOrigin(new_origin)
        assert wcs != wcs3, name+' is not != wcs.shiftOrigin(pos)'
        wcs4 = wcs.local(wcs.origin)
        assert wcs != wcs4, name+' is not != wcs.local()'
        assert wcs4 != wcs, name+' is not != wcs.local() (reverse)'
        world_pos1 = wcs.toWorld(galsim.PositionD(0,0))
        wcs3 = wcs.shiftOrigin(new_origin)
        world_pos2 = wcs3.toWorld(new_origin)
        np.testing.assert_almost_equal(
                world_pos2.distanceTo(world_pos1) / galsim.arcsec, 0, 7,
                'shiftOrigin(new_origin) returned wrong world position')

@timer
def test_wfirst():
    """Test that the deprecated wfirst module works like the new roman module.
    """
    import galsim.roman
    check_dep(__import__, 'galsim.wfirst')

    assert galsim.wfirst.gain == galsim.roman.gain
    assert galsim.wfirst.pixel_scale == galsim.roman.pixel_scale
    assert galsim.wfirst.diameter == galsim.roman.diameter
    assert galsim.wfirst.obscuration == galsim.roman.obscuration
    assert galsim.wfirst.collecting_area == galsim.roman.collecting_area
    assert galsim.wfirst.exptime == galsim.roman.exptime
    assert galsim.wfirst.dark_current == galsim.roman.dark_current
    assert galsim.wfirst.nonlinearity_beta == galsim.roman.nonlinearity_beta
    assert galsim.wfirst.reciprocity_alpha == galsim.roman.reciprocity_alpha
    assert galsim.wfirst.read_noise == galsim.roman.read_noise
    assert galsim.wfirst.n_dithers == galsim.roman.n_dithers
    assert galsim.wfirst.thermal_backgrounds == galsim.roman.thermal_backgrounds
    assert galsim.wfirst.longwave_bands == galsim.roman.longwave_bands
    assert galsim.wfirst.shortwave_bands == galsim.roman.shortwave_bands
    assert galsim.wfirst.pupil_plane_file == galsim.roman.pupil_plane_file
    assert galsim.wfirst.pupil_plane_scale == galsim.roman.pupil_plane_scale
    assert galsim.wfirst.stray_light_fraction == galsim.roman.stray_light_fraction
    np.testing.assert_array_equal(galsim.wfirst.ipc_kernel, galsim.roman.ipc_kernel)
    np.testing.assert_array_equal(galsim.wfirst.persistence_coefficients,
                                  galsim.roman.persistence_coefficients)
    np.testing.assert_array_equal(galsim.wfirst.persistence_fermi_parameters,
                                  galsim.roman.persistence_fermi_parameters)
    assert galsim.wfirst.n_sca == galsim.roman.n_sca
    assert galsim.wfirst.n_pix_tot == galsim.roman.n_pix_tot
    assert galsim.wfirst.n_pix == galsim.roman.n_pix
    assert galsim.wfirst.jitter_rms == galsim.roman.jitter_rms
    assert galsim.wfirst.charge_diffusion == galsim.roman.charge_diffusion

    assert galsim.wfirst.getBandpasses is galsim.roman.getBandpasses
    assert galsim.wfirst.getSkyLevel is galsim.roman.getSkyLevel
    assert galsim.wfirst.getPSF is galsim.roman.getPSF
    assert galsim.wfirst.getWCS is galsim.roman.getWCS
    assert galsim.wfirst.findSCA is galsim.roman.findSCA
    assert galsim.wfirst.allowedPos is galsim.roman.allowedPos
    assert galsim.wfirst.bestPA is galsim.roman.bestPA
    assert galsim.wfirst.convertCenter is galsim.roman.convertCenter
    assert galsim.wfirst.applyNonlinearity is galsim.roman.applyNonlinearity
    assert galsim.wfirst.addReciprocityFailure is galsim.roman.addReciprocityFailure
    assert galsim.wfirst.applyIPC is galsim.roman.applyIPC
    assert galsim.wfirst.applyPersistence is galsim.roman.applyPersistence
    assert galsim.wfirst.allDetectorEffects is galsim.roman.allDetectorEffects
    assert galsim.wfirst.NLfunc is galsim.roman.NLfunc

@timer
def test_roman_psfs(run_slow):
    """Test the deprecated high_accuracy and approximate_struts options.
    """
    import galsim.roman

    test_kwargs = [
        ({ 'approximate_struts':True, 'high_accuracy':False },
         { 'pupil_bin':8 }),
        ({ 'approximate_struts':True, 'high_accuracy':True },
         { 'pupil_bin':4, 'gsparams':galsim.GSParams(folding_threshold=2.e-3) }),
        ({ 'approximate_struts':False, 'high_accuracy':False },
         { 'pupil_bin':4 }),
    ]
    if run_slow:
        test_kwargs.append(
            ({ 'approximate_struts':False, 'high_accuracy':True },
            { 'pupil_bin':1, 'gsparams':galsim.GSParams(folding_threshold=2.e-3) }),
        )

    use_sca = 5
    for kwargs1, kwargs2 in test_kwargs:
        psf1 = check_dep(galsim.roman.getPSF, use_sca, 'Y106', **kwargs1)
        psf2 = galsim.roman.getPSF(use_sca, 'Y106', **kwargs2)
        assert psf1 == psf2

    # Cheat to get coverage of False,True option without spending a long time doing the
    # pupil plane FFT for that one.
    with assert_raises(TypeError):
        check_dep(galsim.roman.getPSF, SCA=use_sca, bandpass='Z087',
                            approximate_struts=False, high_accuracy=True,
                            wavelength='Z087')

@timer
def test_surface_ops():

    # Based on test_sensor.py:test_wavelengths_and_angles, but massively simplified.

    rng = galsim.BaseDeviate(1234)

    fratio = 1.2
    obscuration = 0.2
    assigner = check_dep(galsim.FRatioAngles, fratio, obscuration, rng=rng)

    sed = galsim.SED('CWW_E_ext.sed', 'nm', 'flambda').thin()
    bandpass = galsim.Bandpass('LSST_i.dat', 'nm').thin()
    sampler = check_dep(galsim.WavelengthSampler, sed, bandpass, rng=rng)

    obj = galsim.Gaussian(flux=353, sigma=0.3)
    im = galsim.Image(63,63, scale=1)
    check_dep(obj.drawImage, im, method='phot', surface_ops=[sampler, assigner], rng=rng,
              save_photons=True)

    rng.reset(1234)
    assigner.rng.reset(rng)
    sampler.rng.reset(rng)
    photons = check_dep(obj.makePhot, surface_ops=[sampler, assigner], rng=rng)
    assert photons == im.photons

    rng.reset(1234)
    assigner.rng.reset(rng)
    sampler.rng.reset(rng)
    _, photons2 = check_dep(obj.drawPhot, image=im.copy(), surface_ops=[sampler, assigner], rng=rng)
    assert photons2 == im.photons


@timer
def test_midpoint_basic():
    """Test the basic functionality of the midpt() method.
    """
    # This shouldn't be super accurate, but just make sure it's not really broken.
    x = 0.01*np.arange(1000)
    f = x**2
    result = check_dep(galsim.integ.midpt, f, x)
    expected_val = 10**3./3.
    np.testing.assert_almost_equal(
        result/expected_val, 1.0, decimal=2, verbose=True,
        err_msg='Simple test of midpt() method failed for f(x)=x^2 from 0 to 10')


@timer
def test_trapz_basic():
    """Test the basic functionality of the trapz() method.
    """
    # This shouldn't be super accurate, but just make sure it's not really broken.
    func = lambda x: x**2
    result = check_dep(galsim.integ.trapz, func, 0, 1)
    expected_val = 1.**3./3.
    np.testing.assert_almost_equal(
        result/expected_val, 1.0, decimal=6, verbose=True,
        err_msg='Simple test of trapz() method failed for f(x)=x^2 from 0 to 1')

    result = check_dep(galsim.integ.trapz, func, 0, 1, np.linspace(0, 1, 100000))
    expected_val = 1.**3./3.
    np.testing.assert_almost_equal(
        result/expected_val, 1.0, decimal=6, verbose=True,
        err_msg='Test of trapz() with points failed for f(x)=x^2 from 0 to 1')

    with assert_raises(ValueError):
        check_dep(galsim.integ.trapz, func, 0, 1, points=np.linspace(0, 1.1, 100))
    with assert_raises(ValueError):
        check_dep(galsim.integ.trapz, func, 0.1, 1, points=np.linspace(0, 1, 100))
    with assert_raises(TypeError):
        check_dep(galsim.integ.trapz, func, 0.1, 1, points=2.3)

@timer
def test_hsm_depr():
    hsmp = check_dep(galsim.hsm.HSMParams, max_moment_nsig2=25.0)
    assert hsmp.max_moment_nsig2 == 0.

@timer
def test_photon_array_depr():
    nphotons = 1000

    # First create from scratch
    photon_array = galsim.PhotonArray(nphotons)
    assert len(photon_array.x) == nphotons
    assert len(photon_array.y) == nphotons
    assert len(photon_array.flux) == nphotons
    assert not photon_array.hasAllocatedWavelengths()
    assert not photon_array.hasAllocatedAngles()
    assert not photon_array.hasAllocatedTimes()
    assert not photon_array.hasAllocatedPupil()

    # No deprecation warning using setter
    photon_array.dxdz = 0.17
    assert photon_array.hasAllocatedAngles()
    assert len(photon_array.dxdz) == nphotons
    assert len(photon_array.dydz) == nphotons
    np.testing.assert_array_equal(photon_array.dxdz, 0.17)
    np.testing.assert_array_equal(photon_array.dydz, 0.)

    photon_array.dydz = 0.59
    assert photon_array.hasAllocatedAngles()
    assert len(photon_array.dxdz) == nphotons
    assert len(photon_array.dydz) == nphotons
    np.testing.assert_array_equal(photon_array.dxdz, 0.17)
    np.testing.assert_array_equal(photon_array.dydz, 0.59)

    photon_array.wavelength = 500.
    assert photon_array.hasAllocatedWavelengths()
    assert len(photon_array.wavelength) == nphotons
    np.testing.assert_array_equal(photon_array.wavelength, 500)

    photon_array.pupil_u = 6.0
    assert photon_array.hasAllocatedPupil()
    assert len(photon_array.pupil_u) == nphotons
    assert len(photon_array.pupil_v) == nphotons
    np.testing.assert_array_equal(photon_array.pupil_u, 6.0)
    np.testing.assert_array_equal(photon_array.pupil_v, 0.0)

    photon_array.time = 0.0
    assert photon_array.hasAllocatedTimes()
    assert len(photon_array.time) == nphotons
    np.testing.assert_array_equal(photon_array.time, 0.0)

    # Using the getter is allowed, but deprecated.
    photon_array = galsim.PhotonArray(nphotons)
    dxdz = check_dep(getattr, photon_array, 'dxdz')
    assert photon_array.hasAllocatedAngles()
    assert photon_array.hasAllocatedAngles()
    assert len(photon_array.dxdz) == nphotons
    assert len(photon_array.dydz) == nphotons
    dxdz[:] = 0.17
    np.testing.assert_array_equal(photon_array.dxdz, 0.17)
    np.testing.assert_array_equal(photon_array.dydz, 0.)

    dydz = photon_array.dydz  # Allowed now.
    dydz[:] = 0.59
    np.testing.assert_array_equal(photon_array.dydz, 0.59)

    wave = check_dep(getattr, photon_array, 'wavelength')
    assert photon_array.hasAllocatedWavelengths()
    assert len(photon_array.wavelength) == nphotons
    wave[:] = 500.
    np.testing.assert_array_equal(photon_array.wavelength, 500)

    u = check_dep(getattr, photon_array, 'pupil_u')
    assert photon_array.hasAllocatedPupil()
    assert len(photon_array.pupil_u) == nphotons
    assert len(photon_array.pupil_v) == nphotons
    u[:] = 6.0
    np.testing.assert_array_equal(photon_array.pupil_u, 6.0)
    np.testing.assert_array_equal(photon_array.pupil_v, 0.0)
    v = photon_array.pupil_v
    v[:] = 10.0
    np.testing.assert_array_equal(photon_array.pupil_v, 10.0)

    t = check_dep(getattr, photon_array, 'time')
    assert photon_array.hasAllocatedTimes()
    assert len(photon_array.time) == nphotons
    np.testing.assert_array_equal(photon_array.time, 0.0)
    t[:] = 10
    np.testing.assert_array_equal(photon_array.time, 10.0)

    # For coverage, also need to test the two pair ones in other order.
    photon_array = galsim.PhotonArray(nphotons)
    dydz = check_dep(getattr, photon_array, 'dydz')
    assert photon_array.hasAllocatedAngles()
    assert photon_array.hasAllocatedAngles()
    assert len(photon_array.dxdz) == nphotons
    assert len(photon_array.dydz) == nphotons
    dydz[:] = 0.59
    np.testing.assert_array_equal(photon_array.dxdz, 0.)
    np.testing.assert_array_equal(photon_array.dydz, 0.59)

    dxdz = photon_array.dxdz  # Allowed now.
    dxdz[:] = 0.17
    np.testing.assert_array_equal(photon_array.dxdz, 0.17)

    v = check_dep(getattr, photon_array, 'pupil_v')
    assert photon_array.hasAllocatedPupil()
    assert len(photon_array.pupil_u) == nphotons
    assert len(photon_array.pupil_v) == nphotons
    v[:] = 10.0
    np.testing.assert_array_equal(photon_array.pupil_u, 0.0)
    np.testing.assert_array_equal(photon_array.pupil_v, 10.0)
    u = photon_array.pupil_u
    u[:] = 6.0
    np.testing.assert_array_equal(photon_array.pupil_u, 6.0)

    # Check assignAt
    pa1 = galsim.PhotonArray(50)
    pa1.x = photon_array.x[:50]
    for i in range(50):
        pa1.y[i] = photon_array.y[i]
    pa1.flux[0:50] = photon_array.flux[:50]
    pa1.dxdz = photon_array.dxdz[:50]
    pa1.dydz = photon_array.dydz[:50]
    pa1.pupil_u = photon_array.pupil_u[:50]
    pa1.pupil_v = photon_array.pupil_v[:50]
    pa2 = galsim.PhotonArray(100)
    check_dep(pa2.assignAt, 0, pa1)
    check_dep(pa2.assignAt, 50, pa1)
    np.testing.assert_almost_equal(pa2.x[:50], pa1.x)
    np.testing.assert_almost_equal(pa2.y[:50], pa1.y)
    np.testing.assert_almost_equal(pa2.flux[:50], pa1.flux)
    np.testing.assert_almost_equal(pa2.dxdz[:50], pa1.dxdz)
    np.testing.assert_almost_equal(pa2.dydz[:50], pa1.dydz)
    np.testing.assert_almost_equal(pa2.pupil_u[:50], pa1.pupil_u)
    np.testing.assert_almost_equal(pa2.pupil_v[:50], pa1.pupil_v)
    np.testing.assert_almost_equal(pa2.x[50:], pa1.x)
    np.testing.assert_almost_equal(pa2.y[50:], pa1.y)
    np.testing.assert_almost_equal(pa2.flux[50:], pa1.flux)
    np.testing.assert_almost_equal(pa2.dxdz[50:], pa1.dxdz)
    np.testing.assert_almost_equal(pa2.dydz[50:], pa1.dydz)
    np.testing.assert_almost_equal(pa2.pupil_u[50:], pa1.pupil_u)
    np.testing.assert_almost_equal(pa2.pupil_v[50:], pa1.pupil_v)

    # Error if it doesn't fit.
    with assert_raises(ValueError):
        check_dep(pa2.assignAt, 90, pa1)

@timer
def test_chromatic_flux():
    # This is based on a snippet of test_chromatic_flux in test_chromatic.py.

    bulge_SED = galsim.SED('CWW_E_ext.sed', wave_type='ang', flux_type='flambda')
    star = galsim.Gaussian(fwhm=1e-8) * bulge_SED
    mono_PSF = galsim.Gaussian(half_light_radius=0.8)
    zenith_angle = 20 * galsim.degrees
    bandpass = galsim.Bandpass('LSST_i.dat', 'nm').thin()
    PSF = galsim.ChromaticAtmosphere(mono_PSF, base_wavelength=500,
                                     zenith_angle=zenith_angle)
    PSF = PSF * 1.0
    PSF1 = PSF.interpolate(waves=np.linspace(bandpass.blue_limit, bandpass.red_limit, 30),
                          use_exact_sed=False)

    # Check deprecated use_exact_SED kwarg
    PSF2 = check_dep(PSF.interpolate,
                     waves=np.linspace(bandpass.blue_limit, bandpass.red_limit, 30),
                     use_exact_SED=False)
    assert PSF2 == PSF1

    # Also do this manually with the InterpolatedChromaticObject class
    PSF3 = check_dep(galsim.InterpolatedChromaticObject, PSF,
                     waves=np.linspace(bandpass.blue_limit, bandpass.red_limit, 30),
                     use_exact_SED=False)
    assert PSF3 == PSF1

    # And check deprecated SED attribute.
    sed = check_dep(getattr, PSF, 'SED')
    assert sed == PSF.sed
    sed1 = check_dep(getattr, PSF1, 'SED')
    assert sed1 == PSF1.sed
    sed2 = check_dep(getattr, mono_PSF, 'SED')
    assert sed1 == mono_PSF.sed == galsim.SED(1, 'nm', '1')

@timer
def test_W149():
    # Based on test_config_psf, using old W149 name.

    config = {
        'modules' : ['galsim.roman'],
        'psf' : { 'type' : 'RomanPSF', 'SCA': 4, 'bandpass': 'W149' }
    }

    galsim.config.ImportModules(config)
    psf1 = check_dep(galsim.config.BuildGSObject, config, 'psf')[0]
    psf2 = check_dep(galsim.roman.getPSF, SCA=4, bandpass='W149')
    print('psf1 = ',str(psf1))
    print('psf2 = ',str(psf2))
    assert psf1 == psf2

    config = galsim.config.CleanConfig(config)
    config['image'] = {
        'bandpass' : { 'type' : 'RomanBandpass', 'name' : 'W149' }
    }
    config['psf']['wavelength'] = 985
    config['psf']['pupil_bin'] = 8
    bp = check_dep(galsim.config.BuildBandpass, config['image'], 'bandpass', config)[0]
    config['bandpass'] = bp
    psf1 = check_dep(galsim.config.BuildGSObject, config, 'psf')[0]
    psf2 = check_dep(galsim.roman.getPSF, SCA=4, bandpass='W149', pupil_bin=8, wavelength=985.)
    print('psf1 = ',str(psf1))
    print('psf2 = ',str(psf2))
    assert psf1 == psf2

@timer
def test_photon_array_correlated():

    # This is part of test_convolve in test_photon_array.py
    nphotons = 1000
    obj = galsim.Gaussian(flux=1.7, sigma=2.3)
    rng = galsim.UniformDeviate(1234)
    pa1 = obj.shoot(nphotons, rng)
    rng2 = rng.duplicate()  # Save this state.
    pa2 = obj.shoot(nphotons, rng)

    # If not correlated then convolve is deterministic
    conv_x = pa1.x + pa2.x
    conv_y = pa1.y + pa2.y
    conv_flux = pa1.flux * pa2.flux * nphotons

    np.testing.assert_allclose(np.sum(pa1.flux), 1.7)
    np.testing.assert_allclose(np.sum(pa2.flux), 1.7)
    np.testing.assert_allclose(np.sum(conv_flux), 1.7*1.7)

    np.testing.assert_allclose(np.sum(pa1.x**2)/nphotons, 2.3**2, rtol=0.1)
    np.testing.assert_allclose(np.sum(pa2.x**2)/nphotons, 2.3**2, rtol=0.1)
    np.testing.assert_allclose(np.sum(conv_x**2)/nphotons, 2.*2.3**2, rtol=0.1)

    np.testing.assert_allclose(np.sum(pa1.y**2)/nphotons, 2.3**2, rtol=0.1)
    np.testing.assert_allclose(np.sum(pa2.y**2)/nphotons, 2.3**2, rtol=0.1)
    np.testing.assert_allclose(np.sum(conv_y**2)/nphotons, 2.*2.3**2, rtol=0.1)

    pa3 = galsim.PhotonArray(nphotons)
    pa3.copyFrom(pa1)  # copy from pa1
    pa3.convolve(pa2)
    np.testing.assert_allclose(pa3.x, conv_x)
    np.testing.assert_allclose(pa3.y, conv_y)
    np.testing.assert_allclose(pa3.flux, conv_flux)

    # If one of them is correlated, it is still deterministic.
    pa3.copyFrom(pa1)
    check_dep(pa3.setCorrelated)
    pa3.convolve(pa2)
    np.testing.assert_allclose(pa3.x, conv_x)
    np.testing.assert_allclose(pa3.y, conv_y)
    np.testing.assert_allclose(pa3.flux, conv_flux)

    pa3.copyFrom(pa1)
    check_dep(pa3.setCorrelated, False)
    check_dep(pa2.setCorrelated)
    pa3.convolve(pa2)
    np.testing.assert_allclose(pa3.x, conv_x)
    np.testing.assert_allclose(pa3.y, conv_y)
    np.testing.assert_allclose(pa3.flux, conv_flux)

    # But if both are correlated, then it's not this simple.
    pa3.copyFrom(pa1)
    check_dep(pa3.setCorrelated)
    assert check_dep(pa3.isCorrelated)
    assert check_dep(pa2.isCorrelated)
    pa3.convolve(pa2, rng=rng)
    with assert_raises(AssertionError):
        np.testing.assert_allclose(pa3.x, conv_x)
    with assert_raises(AssertionError):
        np.testing.assert_allclose(pa3.y, conv_y)
    np.testing.assert_allclose(np.sum(pa3.flux), 1.7*1.7)
    np.testing.assert_allclose(np.sum(pa3.x**2)/nphotons, 2*2.3**2, rtol=0.1)
    np.testing.assert_allclose(np.sum(pa3.y**2)/nphotons, 2*2.3**2, rtol=0.1)

    # Can also effect the convolution by treating the psf as a PhotonOp
    pa3.copyFrom(pa1)
    check_dep(pa3.setCorrelated)
    obj.applyTo(pa3, rng=rng2)
    np.testing.assert_allclose(pa3.x, conv_x)
    np.testing.assert_allclose(pa3.y, conv_y)
    np.testing.assert_allclose(pa3.flux, conv_flux)

    # Check the is_corr flag gets set
    assert not check_dep(pa1.isCorrelated)
    pa3 = check_dep(galsim.PhotonArray.fromArrays, pa1.x, pa1.y, pa1.flux, is_corr=True)
    assert check_dep(pa3.isCorrelated)

    # Check toggling is_corr
    assert not check_dep(pa1.isCorrelated)
    check_dep(pa1.setCorrelated)
    assert check_dep(pa1.isCorrelated)
    check_dep(pa1.setCorrelated, False)
    assert not check_dep(pa1.isCorrelated)
    check_dep(pa1.setCorrelated, True)
    assert check_dep(pa1.isCorrelated)


@timer
def test_atredshift():
    """Test the equivalence of obj.atRedshift and the equivalent with SED.atRedshift
    """
    from test_chromatic import disk_SED, bulge_SED, bandpass

    # First simple separable galaxy
    gal = galsim.Sersic(n=3, half_light_radius=1.5)
    gal = gal.shear(g1=0.3, g2=0.2)
    gal1 = gal * bulge_SED
    gal2 = gal * bulge_SED.atRedshift(1.7)

    psf = galsim.Moffat(beta=2.5, half_light_radius=0.3)
    psf = galsim.ChromaticAtmosphere(psf, base_wavelength=500.0, zenith_angle=17*galsim.degrees)

    final1 = galsim.Convolve(check_dep(gal1.atRedshift, 1.7), psf)
    final2 = galsim.Convolve(gal2, psf)
    final3 = galsim.Convolve(check_dep(gal1.expand(lambda w:1.0).atRedshift, 1.7), psf)

    image1 = final1.drawImage(nx=64, ny=64, scale=0.2, bandpass=bandpass)
    image2 = final2.drawImage(nx=64, ny=64, scale=0.2, bandpass=bandpass)
    image3 = final3.drawImage(nx=64, ny=64, scale=0.2, bandpass=bandpass)
    np.testing.assert_allclose(image1.array, image2.array)
    np.testing.assert_allclose(image1.array, image3.array, atol=1.e-5)
    check_pickle(final1)

    # ChromaticSum
    gal1 = gal * bulge_SED + gal.dilate(1.3) * disk_SED
    gal2 = gal * bulge_SED.atRedshift(1.7) + gal.dilate(1.3) * disk_SED.atRedshift(1.7)
    gal3 = check_dep((gal * bulge_SED).atRedshift, 1.7) + \
           check_dep((gal.dilate(1.3) * disk_SED).atRedshift, 1.7)
    final1 = galsim.Convolve(check_dep(gal1.atRedshift, 1.7), psf)
    final2 = galsim.Convolve(gal2, psf)
    final3 = galsim.Convolve(gal3, psf)
    image1 = final1.drawImage(nx=64, ny=64, scale=0.2, bandpass=bandpass)
    image2 = final2.drawImage(nx=64, ny=64, scale=0.2, bandpass=bandpass)
    image3 = final3.drawImage(nx=64, ny=64, scale=0.2, bandpass=bandpass)
    np.testing.assert_allclose(image1.array, image2.array)
    np.testing.assert_allclose(image1.array, image3.array)

    # Probably none of the other Chromatic classes make sense to call atRedshift, so let
    # them use the base class implementation if they do so.
    # Just check ChromaticConvolution as one that doesn't have its own implementation.
    # (This is probably the least implausible use of atRedshift for one of these other classes.)
    smear = galsim.ChromaticObject(galsim.Gaussian(sigma=0.4))
    smear1 = smear.expand(lambda wave: (wave/700)**0.3)
    gal1 = galsim.Convolve(gal * bulge_SED, smear1)
    # Smear is at redshift 1.7, so scale wave by factor of (1+1.7)
    smear2 = smear.expand(lambda wave: (wave/700/2.7)**0.3)
    gal2 = galsim.Convolve(gal * bulge_SED.atRedshift(1.7), smear2)
    final1 = galsim.Convolve(check_dep(gal1.atRedshift, 1.7), psf)
    final2 = galsim.Convolve(gal2, psf)
    image1 = final1.drawImage(nx=64, ny=64, scale=0.2, bandpass=bandpass)
    image2 = final2.drawImage(nx=64, ny=64, scale=0.2, bandpass=bandpass)
    np.testing.assert_allclose(image1.array, image2.array, atol=1.e-6)
    assert gal1.redshift == 0.
    assert check_dep(gal1.atRedshift, 1.7).redshift == 1.7
    assert gal2.redshift == 1.7

    # Finally, if we call atRedshift on a regular GSObject, it doesn't do much.
    gal3 = check_dep(gal.atRedshift, 1.7)
    assert gal == gal3
    # But it does add an attribute, which doesn't obviate equality.
    assert gal3.redshift == 1.7
    assert gal.redshift == 0.

    # One deprecated test from test_chromatic in test_config_image.py.
    # (Simplified to just test the deprecation.)
    config = {
        'gal': {
            'type': 'Exponential',
            'half_light_radius': 0.5,
            'sed': {
                'file_name': 'CWW_E_ext.sed',
                'wave_type': 'Ang',
                'flux_type': 'flambda',
                'norm_flux_density': 1.0,
                'norm_wavelength': 500,
            },
            'redshift': 0.8,
        },
    }
    gal1, _ = check_dep(galsim.config.BuildGSObject, config, 'gal')
    sed = galsim.SED('CWW_E_ext.sed', 'Ang', 'flambda').withFluxDensity(1.0, 500).atRedshift(0.8)
    gal2 = galsim.Exponential(half_light_radius=0.5) * sed
    assert gal1 == gal2

    config = {
        'gal': {
            'type': 'Sum',
            'items': [
                {
                    'type': 'DeVaucouleurs',
                    'half_light_radius': 0.5,
                    'sed': {
                        'file_name': 'CWW_E_ext.sed',
                        'wave_type': 'Ang',
                        'flux_type': 'flambda',
                        'norm_flux_density': 1.0,
                        'norm_wavelength': 500,
                    },
                },
                {
                    'type': 'Exponential',
                    'half_light_radius': 2.0,
                    'sed': {
                        'file_name': 'CWW_Im_ext.sed',
                        'wave_type': 'Ang',
                        'flux_type': 'flambda',
                        'norm_flux_density': 1.0,
                        'norm_wavelength': 500,
                    },
                },
            ],
            'redshift': 0.8,
        },
    }
    gal1, _ = check_dep(galsim.config.BuildGSObject, config, 'gal')
    sed1 = galsim.SED('CWW_E_ext.sed', 'Ang', 'flambda').withFluxDensity(1.0, 500).atRedshift(0.8)
    sed2 = galsim.SED('CWW_Im_ext.sed', 'Ang', 'flambda').withFluxDensity(1.0, 500).atRedshift(0.8)
    gal2 = galsim.DeVaucouleurs(half_light_radius=0.5) * sed1 + \
           galsim.Exponential(half_light_radius=2.0) * sed2
    print(gal1)
    print(gal2)
    print(gal1.obj_list == gal2.obj_list)
    print(gal1.obj_list[0] == gal2.obj_list[0])
    print(gal1.obj_list[0].original == gal2.obj_list[0].original)
    print(gal1.obj_list[0].sed == gal2.obj_list[0].sed)
    print(gal1.obj_list[1] == gal2.obj_list[1])
    print(gal1._gsparams == gal2._gsparams)
    print(gal1._propagate_gsparams == gal2._propagate_gsparams)
    assert gal1 == gal2


@timer
def test_save_photons():
    # These are the deprecated cases in test_save_photons.
    from test_chromatic import disk_SED, bulge_SED

    rng = galsim.BaseDeviate(1234)
    star_sed = galsim.SED('vega.txt', wave_type="nm", flux_type="fphotons")
    bandpass = galsim.Bandpass("LSST_r.dat", wave_type="nm")

    airy = galsim.ChromaticAiry(lam=500, diam=8)
    optical = galsim.ChromaticOpticalPSF(lam=500, diam=8, defocus=0.2, obscuration=0.3)
    disk = galsim.Exponential(half_light_radius=0.5).shear(g1=0.4, g2=0.2)
    bulge = galsim.Sersic(n=3, half_light_radius=0.3)

    objs = [
        check_dep((disk * disk_SED).atRedshift, 1.1),
        check_dep((optical * star_sed).atRedshift, 1.1),
        check_dep((bulge * bulge_SED + disk * disk_SED).atRedshift, 0.5),
        check_dep((airy * star_sed).expand(lambda w: (w/500)**0.0).atRedshift, 0.2),
        check_dep(galsim.ChromaticTransformation, disk, flux_ratio=disk_SED, redshift=1.1),
    ]

    flux = 1000
    for obj in objs:
        print('obj = ',obj)
        obj = obj.withFlux(flux, bandpass)
        image = obj.drawImage(bandpass=bandpass, method="phot",
                              n_photons=flux, save_photons=True,
                              scale=0.05, nx=32, ny=32, rng=rng)
        assert hasattr(image, 'photons')
        assert len(image.photons) == flux
        print(np.sum(image.photons.flux))
        # Note: tolerance is quite loose, since profiles that use InterpolatedImage can have
        # negative flux photons, which then don't necessarily sum to the right value.
        # Only the expectation value is right, and we're not shooting many photons here.
        assert np.allclose(np.sum(image.photons.flux), flux, rtol=0.1)

        # Sometimes there is a different path when n_photons is not given, so check that too.
        image = obj.drawImage(bandpass=bandpass, method="phot",
                              save_photons=True,
                              scale=0.05, nx=32, ny=32, rng=rng)
        assert hasattr(image, 'photons')
        print(np.sum(image.photons.flux))
        assert np.allclose(np.sum(image.photons.flux), flux, rtol=0.1)
        repr(obj)


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#


# These have the basic config functionality that gets imported into galsim.config scope
from .process import *
from .input import *
from .output import *
from .extra import *
from .image import *
from .stamp import *
from .noise import *
from .wcs import *
from .bandpass import *
from .sed import *
from .photon_ops import *
from .sensor import *
from .gsobject import *
from .value import *
from .value_eval import eval_base_variables
from .util import *

# These implement specific types and features that get registered into the main config
# apparatus.  The functions themselves are not available at galsim.config scope.
from . import output_multifits
from . import output_datacube

from . import input_fitsheader
from . import input_real
from . import input_cosmos
from . import input_nfw
from . import input_powerspectrum
from . import input_image

from . import extra_psf
from . import extra_weight
from . import extra_badpix
from . import extra_truth

from . import image_scattered
from . import image_tiled

from . import stamp_ring

from . import value_random
from . import value_eval



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'Catalog', 'Dict', 'OutputCatalog' ]

import os
import numpy as np
import pickle
import yaml
import json

from .errors import GalSimValueError, GalSimKeyError, GalSimIndexError
from ._utilities import lazy_property
from ._pyfits import pyfits
from .angle import Angle
from .position import PositionD, PositionI
from .shear import Shear

class Catalog:
    """A class storing the data from an input catalog.

    Each row corresponds to a different object to be built, and each column stores some item of
    information about that object (e.g. flux or half_light_radius).

    Parameters:
        file_name:  Filename of the input catalog. (Required)
        dir:        Optionally a directory name can be provided if ``file_name`` does not
                    already include it.
        file_type:  Either 'ASCII' or 'FITS'.  If None, infer from ``file_name`` ending.
                    [default: None]
        comments:   The character used to indicate the start of a comment in an
                    ASCII catalog.  [default: '#']
        hdu:        Which hdu to use for FITS files.  [default: 1]

    After construction, the following attributes are available:

    Attributes:
        nobjects:   The number of objects in the catalog.
        ncols:      The number of columns in the catalog.
        isfits:     Whether the catalog is a fits catalog.
        names:      For a fits catalog, the valid column names.
    """
    _req_params = { 'file_name' : str }
    _opt_params = { 'dir' : str , 'file_type' : str , 'comments' : str , 'hdu' : int }

    def __init__(self, file_name, dir=None, file_type=None, comments='#', hdu=1):

        # First build full file_name
        self.file_name = file_name.strip()
        if dir is not None:
            self.file_name = os.path.join(dir,self.file_name)

        if file_type is None:
            name, ext = os.path.splitext(file_name)
            if ext.lower().startswith('.fit'):
                file_type = 'FITS'
            else:
                file_type = 'ASCII'
        file_type = file_type.upper()
        if file_type not in ('FITS', 'ASCII'):
            raise GalSimValueError("file_type must be either FITS or ASCII if specified.",
                                   file_type, ('FITS', 'ASCII'))
        self.file_type = file_type
        if comments == '': comments = None  # loadtxt actually wants None, not ''
        self.comments = comments
        self.hdu = hdu

        if file_type == 'FITS':
            self.readFits()
        else:  # file_type == 'ASCII':
            self.readAscii()

    def finishRead(self):
        if self.file_type == 'ASCII':
            self.finishReadAscii()
        # FITS is already finished.

    # A couple things only get set for ASCII if finishRead has been run.
    @lazy_property
    def data(self):
        self.finishRead()
        return self._data

    @lazy_property
    def ncols(self):
        self.finishRead()
        return self._ncols

    # When we make a proxy of this class (cf. galsim/config/stamp.py), the attributes
    # don't get proxied.  Only callable methods are.  So make method versions of these.
    def getNObjects(self) : return self.nobjects
    def isFits(self) : return self.isfits
    def __len__(self) : return self.nobjects

    def readAscii(self):
        """Read in an input catalog from an ASCII file.
        """
        if self.comments is not None and len(self.comments) > 1:
            raise GalSimValueError('Invalid comments character', self.comments)

        # In the first pass, just figure out nobjects.  We may not need to go past this.
        # See the script devel/testlinecounting.py that tests several possibilities.
        # An even faster version using buffering is possible although it requires some care
        # around edge cases, so we use this one instead, which is "correct by inspection".
        with open(self.file_name) as f:
            if self.comments is not None:
                c = self.comments[0]
                self.nobjects = sum(1 for line in f if line[0] != c)
            else:  # comments == None.  No comments.
                self.nobjects = sum(1 for line in f)
        self.isfits = False

    def finishReadAscii(self):
        # Read in the data using the numpy convenience function
        # Note: we leave the data as str, rather than convert to float, so that if
        # we have any str fields, they don't give an error here.  They'll only give an
        # error if one tries to convert them to float at some point.
        self._data = np.loadtxt(self.file_name, comments=self.comments, dtype=bytes, ndmin=2)
        # Convert the bytes to str.  For Py2, this is a no op.
        self._data = self._data.astype(str)

        assert self.nobjects == self._data.shape[0]
        self._ncols = self._data.shape[1]

    def readFits(self):
        """Read in an input catalog from a FITS file.
        """
        with pyfits.open(self.file_name) as fits:
            self._data = fits[self.hdu].data.copy()
        self.names = self._data.columns.names
        self.nobjects = len(self._data)
        self._ncols = len(self.names)
        self.isfits = True

    def get(self, index, col):
        """Return the data for the given ``index`` and ``col`` in its native type.

        For ASCII catalogs, ``col`` is the column number.
        For FITS catalogs, ``col`` is a string giving the name of the column in the FITS table.

        Also, for ASCII catalogs, the "native type" is always str.  For FITS catalogs, it is
        whatever type is specified for each field in the binary table.
        """
        if self.isfits:
            if col not in self.names:
                raise GalSimKeyError("Column is invalid for catalog %s"%self.file_name, col)
            if not isinstance(index, int):
                raise GalSimIndexError("Index must be an int for catalog %s"%self.file_name, index)
            if index < 0 or index >= self.nobjects:
                raise GalSimIndexError("Index is invalid for catalog %s"%self.file_name, index)
            return self.data[col][index]
        else:
            if not isinstance(col, int):
                raise GalSimIndexError("Column must an int for ASCII catalog %s"%self.file_name,
                                       col)
            if col < 0 or col >= self.ncols:
                raise GalSimIndexError("Column is invalid for catalog %s"%self.file_name, col)
            if not isinstance(index, int):
                raise GalSimIndexError("Index must be an int for catalog %s"%self.file_name, index)
            if index < 0 or index >= self.nobjects:
                raise GalSimIndexError("Index is invalid for catalog %s"%self.file_name, col)
            return self.data[index, col]

    def getFloat(self, index, col):
        """Return the data for the given ``index`` and ``col`` as a float if possible
        """
        return float(self.get(index,col))

    def getInt(self, index, col):
        """Return the data for the given ``index`` and ``col`` as an int if possible
        """
        return int(self.get(index,col))

    def __repr__(self):
        s = "galsim.Catalog(file_name=%r, file_type=%r"%(self.file_name, self.file_type)
        if self.comments != '#': s += ', comments=%r'%self.comments
        if self.hdu != 1: s += ', hdu=%r'%self.hdu
        s += ')'
        return s

    def __str__(self): return "galsim.Catalog(file_name=%r)"%self.file_name

    def __eq__(self, other): return self is other or repr(self) == repr(other)
    def __ne__(self, other): return not self.__eq__(other)
    def __hash__(self): return hash(repr(self))


class Dict:
    """A class that reads a python dict from a file.

    After construction, it behaves like a regular python dict, with one exception.
    In order to facilitate getting values in a hierarchy of fields, we allow the '.'
    character to chain keys together for the get() method.  So,::

        >>> d.get('noise.properties.variance')

    is expanded into::

        >>> d['noise']['properties']['variance']

    Furthermore, if a "key" is really an integer, then it is used as such, which accesses
    the corresponding element in a list.  e.g.::

        >>> d.get('noise_models.2.variance')

    is equivalent to::

        >>> d['noise_models'][2]['variance']

    This makes it much easier to access arbitrary elements within parameter files.

    Caveat: The above prescription means that an element whose key really has a '.' in it
    won't be accessed correctly.  This is probably a rare occurrence, but the workaround is
    to set ``key_split`` to a different character or string and use that to chain the keys.


    Parameters:
        file_name:  Filename storing the dict.
        dir:        Optionally a directory name can be provided if ``file_name`` does not
                    already include it. [default: None]
        file_type:  Options are 'Pickle', 'YAML', or 'JSON' or None.  If None, infer from
                    ``file_name`` extension ('.p*', '.y*', '.j*' respectively).
                    [default: None]
        key_split:  The character (or string) to use to split chained keys.  (cf. the
                    description of this feature above.)  [default: '.']
    """
    _req_params = { 'file_name' : str }
    _opt_params = { 'dir' : str , 'file_type' : str, 'key_split' : str }
    _single_params = []
    _takes_rng = False

    def __init__(self, file_name, dir=None, file_type=None, key_split='.'):

        # First build full file_name
        self.file_name = file_name.strip()
        if dir is not None:
            self.file_name = os.path.join(dir,self.file_name)

        if file_type is None:
            name, ext = os.path.splitext(self.file_name)
            if ext.lower().startswith('.p'):
                file_type = 'PICKLE'
            elif ext.lower().startswith('.y'):
                file_type = 'YAML'
            elif ext.lower().startswith('.j'):
                file_type = 'JSON'
            else:
                raise GalSimValueError('Unable to determine file_type from file_name ending',
                                       file_name, ('*.p*', '*.y*', '*.j*'))

        file_type = file_type.upper()
        if file_type not in ('PICKLE','YAML','JSON'):
            raise GalSimValueError("Invalid file_type", file_type, ('Pickle', 'YAML', 'JSON'))
        self.file_type = file_type

        self.key_split = key_split

        if file_type == 'PICKLE':
            with open(self.file_name, 'rb') as f:
                self.dict = pickle.load(f)
        elif file_type == 'YAML':
            with open(self.file_name, 'r') as f:
                self.dict = yaml.safe_load(f)
        else:  # JSON
            with open(self.file_name, 'r') as f:
                self.dict = json.load(f)

    def get(self, key, default=None):
        # Make a list of keys according to our key_split parameter
        chain = key.split(self.key_split)
        d = self.dict
        while len(chain):
            k = chain.pop(0)

            # Try to convert to an integer:
            try: k = int(k)
            except ValueError: pass

            # If there are more keys, just set d to the next in the chanin.
            if chain: d = d[k]
            # Otherwise, return the result.
            else:
                if k not in d and default is None:
                    raise GalSimKeyError("key not found in dictionary.",key)
                return d.get(k,default)
        raise GalSimKeyError("Invalid key given to Dict.get()",key)

    # The rest of the functions are typical non-mutating functions for a dict, for which we just
    # pass the request along to self.dict.
    def __len__(self):
        return len(self.dict)

    def __getitem__(self, key):
        return self.dict[key]

    def __contains__(self, key):
        return key in self.dict

    def __iter__(self):
        return self.dict.__iter__()

    def keys(self):
        return self.dict.keys()

    def values(self):
        return self.dict.values()

    def items(self):
        return self.dict.items()

    def iterkeys(self):
        return self.keys()

    def itervalues(self):
        return self.values()

    def iteritems(self):
        return self.items()

    def __repr__(self):
        s = "galsim.Dict(file_name=%r, file_type=%r"%(self.file_name, self.file_type)
        if self.key_split != '.':
            s += ', key_split=%r'%self.key_split
        s += ')'
        return s

    def __str__(self): return "galsim.Dict(file_name=%r)"%self.file_name

    def __eq__(self, other): return self is other or repr(self) == repr(other)
    def __ne__(self, other): return not self.__eq__(other)
    def __hash__(self): return hash(repr(self))



class OutputCatalog:
    """A class for building up a catalog for output, typically storing truth information
    about a simulation.

    Each row corresponds to a different object, and each column stores some item of
    information about that object (e.g. flux or half_light_radius).

    Note: no type checking is done when the data are added in addRow().  It is up to
    the user to make sure that the values added for each row are compatible with the
    types given here in the ``types`` parameter.

    Parameters:
        names:      A list of names for the output columns.
        types:      A list of types for the output columns. [default: None, which assumes all
                    columns are float]

    After construction, the following attributes are available:

    Attributes:
        nobjects:       The number of objects so far in the catalog.
        ncols:          The number of columns in the catalog.
        names:          The names of the columns.
        types:          The types of the columns.
        rows:           The rows of data that have been accumulated so far.
    """
    # Watch out for this "Gotcha".  Using _rows=[] as the default argument doesn't work!
    # https://pythonconquerstheuniverse.wordpress.com/2012/02/15/mutable-default-arguments/
    def __init__(self, names, types=None, _rows=(), _sort_keys=()):
        self.names = names
        if types is None:
            self.types = [ float for i in names ]
        else:
            self.types = types
        self.rows = list(_rows)
        self.sort_keys = list(_sort_keys)

    @property
    def nobjects(self):
        """The number of objects in the `OutputCatalog`.
        """
        return len(self.rows)
    @property
    def ncols(self):
        """The number of columns in the `OutputCatalog`.
        """
        return len(self.names)
    def __len__(self): return self.nobjects

    # Again, when we use this through a proxy, we need getters for the attributes.
    def getNames(self):
        """Equivalent to sef.names."""
        return self.names
    def getTypes(self):
        """Equivalent to sef.types."""
        return self.types
    def setTypes(self, types):
        """Equivalent to sef.types = types."""
        self.types = types
    def getNObjects(self):
        """Equivalent to sef.nobjects."""
        return self.nobjects
    def getNCols(self):
        """Equivalent to sef.ncols."""
        return self.ncols

    def addRow(self, row, sort_key=None):
        """Add a row of data to the catalog.

        Warning: no type checking is done at this point.  If the values in the row do not
        match the column types, you may get an error when writing, or you may lose precision,
        depending on the nature of the mismatch.

        Parameters:
            row:        A list with one item per column in the same order as the names list.
            sort_key:   If the rows may be added out of order, you can provide a sort_key,
                        which will be used at the end to re-sort the rows.
        """
        if len(row) != self.ncols:
            raise GalSimValueError("Length of row does not match the number of columns = %d"%(
                                   self.ncols), len(row))
        self.rows.append(tuple(row))
        if sort_key is None:
            self.sort_keys.append(self.nobjects)
        else:
            self.sort_keys.append(sort_key)

    def write(self, file_name, dir=None, file_type=None, prec=8):
        """Write the catalog to a file.

        Parameters:
            file_name:  The name of the file to write to.
            dir:        Optionally a directory name can be provided if ``file_name`` does not
                        already include it. [default: None]
            file_type:  Which kind of file to write to. [default: determine from the file_name
                        extension]
            prec:       Output precision for ASCII. [default: 8]
        """
        if dir is not None:
            file_name = os.path.join(dir,file_name)

        # Figure out which file type the catalog is
        if file_type is None:
            name, ext = os.path.splitext(file_name)
            if ext.lower().startswith('.fit'):
                file_type = 'FITS'
            else:
                file_type = 'ASCII'
        file_type = file_type.upper()
        if file_type not in ('FITS', 'ASCII'):
            raise GalSimValueError("Invalid file_type.", file_type, ('FITS', 'ASCII'))

        if file_type == 'FITS':
            self.writeFits(file_name)
        else:  # file_type == 'ASCII':
            self.writeAscii(file_name, prec)

    def makeData(self):
        """Returns a numpy array of the data as it should be written to an output file.
        """

        cols = zip(*self.rows)

        dtypes = []
        new_cols = []
        for col, name, t in zip(cols, self.names, self.types):
            name = str(name)  # numpy will barf if the name is a unicode string
            dt = np.dtype(t) # just used to categorize the type into int, float, str
            if dt.kind in np.typecodes['AllInteger']:
                dtypes.append( (name, int) )
                new_cols.append(col)
            elif dt.kind in np.typecodes['AllFloat']:
                dtypes.append( (name, float) )
                new_cols.append(col)
            elif t == Angle:
                dtypes.append( (name + ".rad", float) )
                new_cols.append( [ val.rad for val in col ] )
            elif t == PositionI:
                dtypes.append( (name + ".x", int) )
                dtypes.append( (name + ".y", int) )
                new_cols.append( [ val.x for val in col ] )
                new_cols.append( [ val.y for val in col ] )
            elif t == PositionD:
                dtypes.append( (name + ".x", float) )
                dtypes.append( (name + ".y", float) )
                new_cols.append( [ val.x for val in col ] )
                new_cols.append( [ val.y for val in col ] )
            elif t == Shear:
                dtypes.append( (name + ".g1", float) )
                dtypes.append( (name + ".g2", float) )
                new_cols.append( [ val.g1 for val in col ] )
                new_cols.append( [ val.g2 for val in col ] )
            else:
                col = [ str(s).encode() for s in col ]
                maxlen = np.max([ len(s) for s in col ])
                dtypes.append( (name, str, maxlen) )
                new_cols.append(col)

        data = np.array(list(zip(*new_cols)), dtype=dtypes)

        sort_index = np.argsort(self.sort_keys)
        data = data[sort_index]

        return data

    def writeAscii(self, file_name, prec=8):
        """Write catalog to an ASCII file.

        Parameters:
            file_name:  The name of the file to write to.
            prec:       Output precision for floats. [default: 8]
        """
        data = self.makeData()

        width = prec+8
        header_form = ""
        for i in range(len(data.dtype.names)):
            header_form += "{%d:^%d} "%(i,width)
        header = header_form.format(*data.dtype.names)

        fmt = []
        for name in data.dtype.names:
            dt = data.dtype[name]
            if dt.kind in np.typecodes['AllInteger']:
                fmt.append('%%%dd'%(width))
            elif dt.kind in np.typecodes['AllFloat']:
                fmt.append('%%%d.%de'%(width,prec))
            else:
                fmt.append('%%%ds'%(width))

        np.savetxt(file_name, data, fmt=fmt, header=header)

    def writeFits(self, file_name):
        """Write catalog to a FITS file.

        Parameters:
            file_name:  The name of the file to write to.
        """
        from .fits import writeFile
        tbhdu = self.writeFitsHdu()
        writeFile(file_name, tbhdu)

    def writeFitsHdu(self):
        """Write catalog to a FITS hdu.

        Returns:
            an HDU with the FITS binary table of the catalog.
        """
        # Note to developers: Because of problems with pickling in older pyfits versions, this
        # code is duplicated in galsim/config/extra_truth.py, BuildTruthHDU.  If you change
        # this function, you should update BuildTruthHDU as well.

        data = self.makeData()

        cols = []
        for name in data.dtype.names:
            dt = data.dtype[name]
            if dt.kind in np.typecodes['AllInteger']:
                cols.append(pyfits.Column(name=name, format='J', array=data[name]))
            elif dt.kind in np.typecodes['AllFloat']:
                cols.append(pyfits.Column(name=name, format='D', array=data[name]))
            else:
                cols.append(pyfits.Column(name=name, format='%dA'%dt.itemsize, array=data[name]))

        cols = pyfits.ColDefs(cols)
        tbhdu = pyfits.BinTableHDU.from_columns(cols)
        return tbhdu

    def __repr__(self):
        def make_type_str(t):
            s = repr(t)
            if s[1:6] == 'class': return s[8:-2]
            else: return s
        type_str = "( " + ", ".join([ make_type_str(t) for t in self.types ]) + " )"
        return "galsim.OutputCatalog(names=%r, types=%s, _rows=%r)"%(
                self.names, type_str, self.rows)

    def __str__(self):
        return "galsim.OutputCatalog(names=%r)"%self.names

    def __eq__(self, other): return self is other or repr(self) == repr(other)
    def __ne__(self, other): return not self.__eq__(other)
    def __hash__(self): return hash(repr(self))



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np

from . import _galsim
from .image import Image, ImageD, ImageCD
from .bounds import BoundsI
from .errors import GalSimValueError, convert_cpp_errors

def fft2(a, shift_in=False, shift_out=False):
    """Compute the 2-dimensional discrete Fourier Transform.

    For valid inputs, the result is equivalent to numpy.fft.fft2(a), but usually faster.::

        >>> ka1 = numpy.fft.fft2(a)
        >>> ka2 = galsim.fft.fft2(a)

    Restrictions on this version vs the numpy version:

        - The input array must be 2-dimensional.
        - The size in each direction must be even. (Ideally 2^k or 3*2^k for speed, but this is
          not required.)
        - If it has a real dtype, it will be coerced to numpy.float64.
        - If it has a complex dtype, it will be coerced to numpy.complex128.

    The returned array will be complex with dtype numpy.complex128.

    If shift_in is True, then this is equivalent to applying numpy.fft.fftshift to the input.::

        >>> ka1 = numpy.fft.fft2(numpy.fft.fftshift(a))
        >>> ka2 = galsim.fft.fft2(a, shift_in=True)

    If shift_out is True, then this is equivalent to applying numpy.fft.fftshift to the output.::

        >>> ka1 = numpy.fft.fftshift(numpy.fft.fft2(a))
        >>> ka2 = galsim.fft.fft2(a, shift_out=True)

    Parameters:
        a:          The input array to be transformed
        shift_in:   Whether to shift the input array so that the center is moved to (0,0).
                    [default: False]
        shift_out:  Whether to shift the output array so that the center is moved to (0,0).
                    [default: False]

    Returns:
        a complex numpy array
    """
    s = a.shape
    if len(s) != 2:
        raise GalSimValueError("Input array must be 2D.",s)
    M, N = s
    Mo2 = M // 2
    No2 = N // 2

    if M != Mo2*2 or N != No2*2:
        raise GalSimValueError("Input array must have even sizes.",s)

    if a.dtype.kind == 'c':
        a = a.astype(np.complex128, copy=False)
        xim = ImageCD(a, xmin = -No2, ymin = -Mo2)
        kim = ImageCD(BoundsI(-No2,No2-1,-Mo2,Mo2-1))
        with convert_cpp_errors():
            _galsim.cfft(xim._image, kim._image, False, shift_in, shift_out)
        kar = kim.array
    else:
        a = a.astype(np.float64, copy=False)
        xim = ImageD(a, xmin = -No2, ymin = -Mo2)

        # This works, but it's a bit slower.
        #kim = ImageCD(BoundsI(-No2,No2-1,-Mo2,Mo2-1))
        #_galsim.cfft(xim._image, kim._image, False, shift_in, shift_out)
        #kar = kim.array

        # Faster to start with rfft2 version
        rkim = ImageCD(BoundsI(0,No2,-Mo2,Mo2-1))
        with convert_cpp_errors():
            _galsim.rfft(xim._image, rkim._image, shift_in, shift_out)
        # This only returns kx >= 0.  Fill out the full image.
        kar = np.empty( (M,N), dtype=np.complex128)
        rkar = rkim.array
        if shift_out:
            kar[:,No2:N] = rkar[:,0:No2]
            kar[0,0:No2] = rkar[0,No2:0:-1].conjugate()
            kar[1:Mo2,0:No2] = rkar[M-1:Mo2:-1,No2:0:-1].conjugate()
            kar[Mo2:M,0:No2] = rkar[Mo2:0:-1,No2:0:-1].conjugate()
        else:
            kar[:,0:No2] = rkar[:,0:No2]
            kar[0,No2:N] = rkar[0,No2:0:-1].conjugate()
            kar[1:M,No2:N] = rkar[M-1:0:-1,No2:0:-1].conjugate()
    return kar


def ifft2(a, shift_in=False, shift_out=False):
    """Compute the 2-dimensional inverse discrete Fourier Transform.

    For valid inputs, the result is equivalent to numpy.fft.ifft2(a), but usually faster.::

        >>> a1 = numpy.fft.ifft2(ka)
        >>> a2 = galsim.fft.ifft2(ka)

    Restrictions on this version vs the numpy version:

        - The array must be 2-dimensional.
        - The size in each direction must be even. (Ideally 2^k or 3*2^k for speed, but this is
          not required.)
        - The array is assumed to be Hermitian, which means the k values with kx<0 are assumed
          to be equal to the conjuate of their inverse.  This will always be the case if
          a is an output of fft2 (with a real input array).  i.e.

          - for kx >= N/2, ky > 0: a[ky, kx] == a[N-ky, N-kx].conjugate()
          - for kx >= N/2, ky = 0: a[0, kx] == a[0, N-kx].conjugate()

          Only the elements a[:,0:N/2+1] are accessed by this function.
        - If it has a real dtype, it will be coerced to numpy.float64.
        - If it has a complex dtype, it will be coerced to numpy.complex128.

    The returned array will be complex with dtype numpy.complex128.

    If shift_in is True, then this is equivalent to applying numpy.fft.fftshift to the input::

        >>> a1 = numpy.fft.ifft2(numpy.fft.fftshift(ka))
        >>> a2 = galsim.fft.ifft2(ka, shift_in=True)

    If shift_out is True, then this is equivalent to applying numpy.fft.fftshift to the output::

        >>> a1 = numpy.fft.fftshift(numpy.fft.ifft2(ka))
        >>> a2 = galsim.fft.ifft2(ka, shift_out=True)

    Parameters:
        a:          The input array to be transformed
        shift_in:   Whether to shift the input array so that the center is moved to (0,0).
                    [default: False]
        shift_out:  Whether to shift the output array so that the center is moved to (0,0).
                    [default: False]

    Returns:
        a complex numpy array
    """
    s = a.shape
    if len(s) != 2:
        raise GalSimValueError("Input array must be 2D.",s)
    M,N = s
    Mo2 = M // 2
    No2 = N // 2

    if M != Mo2*2 or N != No2*2:
        raise GalSimValueError("Input array must have even sizes.",s)

    if a.dtype.kind == 'c':
        a = a.astype(np.complex128, copy=False)
        kim = ImageCD(a, xmin = -No2, ymin = -Mo2)
    else:
        a = a.astype(np.float64, copy=False)
        kim = ImageD(a, xmin = -No2, ymin = -Mo2)
    xim = ImageCD(BoundsI(-No2,No2-1,-Mo2,Mo2-1))
    with convert_cpp_errors():
        _galsim.cfft(kim._image, xim._image, True, shift_in, shift_out)
    return xim.array


def rfft2(a, shift_in=False, shift_out=False):
    """Compute the one-dimensional discrete Fourier Transform for real input.

    For valid inputs, the result is equivalent to numpy.fft.rfft2(a), but usually faster.::

        >>> ka1 = numpy.fft.rfft2(a)
        >>> ka2 = galsim.fft.rfft2(a)

    Restrictions on this version vs the numpy version:

        - The input array must be 2-dimensional.
        - If it does not have dtype numpy.float64, it will be coerced to numpy.float64.
        - The size in each direction must be even. (Ideally 2^k or 3*2^k for speed, but this is
          not required.)

    The returned array will be complex with dtype numpy.complex128.

    If shift_in is True, then this is equivalent to applying numpy.fft.fftshift to the input.::

        >>> ka1 = numpy.fft.rfft2(numpy.fft.fftshift(a))
        >>> ka2 = galsim.fft.rfft2(a, shift_in=True)

    If shift_out is True, then this is equivalent to applying numpy.fft.fftshift to the output.::

        >>> ka1 = numpy.fft.fftshift(numpy.fft.rfft2(a),axes=(0,))
        >>> ka2 = galsim.fft.rfft2(a, shift_out=True)

    Parameters:
        a:          The input array to be transformed
        shift_in:   Whether to shift the input array so that the center is moved to (0,0).
                    [default: False]
        shift_out:  Whether to shift the output array so that the center is moved to (0,0).
                    [default: False]

    Returns:
        a complex numpy array
    """
    s = a.shape
    if len(s) != 2:
        raise GalSimValueError("Input array must be 2D.",s)
    M,N = s
    Mo2 = M // 2
    No2 = N // 2

    if M != Mo2*2 or N != No2*2:
        raise GalSimValueError("Input array must have even sizes.",s)

    a = a.astype(np.float64, copy=False)
    xim = ImageD(a, xmin = -No2, ymin = -Mo2)
    kim = ImageCD(BoundsI(0,No2,-Mo2,Mo2-1))
    with convert_cpp_errors():
        _galsim.rfft(xim._image, kim._image, shift_in, shift_out)
    return kim.array


def irfft2(a, shift_in=False, shift_out=False):
    """Compute the 2-dimensional inverse FFT of a real array.

    For valid inputs, the result is equivalent to numpy.fft.irfft2(a), but usually faster.::

        >>> a1 = numpy.fft.irfft2(ka)
        >>> a2 = galsim.fft.irfft2(ka)

    Restrictions on this version vs the numpy version:

        - The array must be 2-dimensional.
        - If it does not have dtype numpy.complex128, it will be coerced to numpy.complex128.
        - It must have shape (M, N/2+1).
        - The size M must be even. (Ideally 2^k or 3*2^k for speed, but this is not required.)

    The returned array will be real with dtype numpy.float64.

    If shift_in is True, then this is equivalent to applying numpy.fft.fftshift to the input.::

        >>> a1 = numpy.fft.irfft2(numpy.fft.fftshift(a, axes=(0,)))
        >>> a2 = galsim.fft.irfft2(a, shift_in=True)

    If shift_out is True, then this is equivalent to applying numpy.fft.fftshift to the output.::

        >>> a1 = numpy.fft.fftshift(numpy.fft.irfft2(a))
        >>> a2 = galsim.fft.irfft2(a, shift_out=True)

    Parameters:
        a:          The input array to be transformed
        shift_in:   Whether to shift the input array so that the center is moved to (0,0).
                    [default: False]
        shift_out:  Whether to shift the output array so that the center is moved to (0,0).
                    [default: False]

    Returns:
        a real numpy array
    """
    s = a.shape
    if len(s) != 2:
        raise GalSimValueError("Input array must be 2D.",s)
    M,No2 = s
    No2 -= 1  # s is (M,No2+1)
    Mo2 = M // 2

    if M != Mo2*2:
        raise GalSimValueError("Input array must have even sizes.",s)

    a = a.astype(np.complex128, copy=False)
    kim = ImageCD(a, xmin = 0, ymin = -Mo2)
    xim = ImageD(BoundsI(-No2,No2+1,-Mo2,Mo2-1))
    with convert_cpp_errors():
        _galsim.irfft(kim._image, xim._image, shift_in, shift_out)
    xim = xim.subImage(BoundsI(-No2,No2-1,-Mo2,Mo2-1))
    return xim.array





# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#
import numpy as np
import re
from astropy.units import Quantity, Unit

from .util import PropagateIndexKeyRNGNum
from .value import GetCurrentValue, GetAllParams, RegisterValueType
from .value import standard_ignore
from .process import ImportModules
from ..errors import GalSimConfigError
from ..angle import Angle
from ..position import PositionD
from ..celestial import CelestialCoord
from ..shear import Shear
from ..table import LookupTable

# This file handles the parsing for the special Eval type.
letter_codes = {
    'f' : float,
    'i' : int,
    'b' : bool,
    's' : str,
    'a' : Angle,
    'p' : PositionD,
    'c' : CelestialCoord,
    'g' : Shear,
    't' : LookupTable,
    'd' : dict,
    'l' : list,
    'x' : None,
    'q' : Quantity,
    'u' : Unit,
}


def _type_by_letter(key):
    if len(key) < 2:
        raise GalSimConfigError("Invalid user-defined variable %r"%key)
    letter = key[0]
    if letter in letter_codes.keys():
        return letter_codes[letter]
    else:
        raise GalSimConfigError(
            "Invalid Eval variable: %s (starts with an invalid letter)"%key)

eval_base_variables = [ 'image_pos', 'world_pos', 'image_center', 'image_origin', 'image_bounds',
                        'image_xsize', 'image_ysize', 'stamp_xsize', 'stamp_ysize', 'pixel_scale',
                        'wcs', 'rng', 'file_num', 'image_num', 'obj_num', 'start_obj_num',
                        'world_center', 'sky_pos', 'uv_pos', 'bandpass' ]

eval_ignore = ['str','_fn'] + standard_ignore

def _isWordInString(w, s):
    # Return if a given word is in the given string.
    # Note, this specifically looks for the whole word. e.g. if w = 'yes' and s = 'eyestrain',
    # then `w in s` returns True, but `_isWordInString(w,s)` returns False.
    # cf. https://stackoverflow.com/questions/5319922/python-check-if-word-is-in-a-string
    return re.search(r'\b({0})\b'.format(w),s) is not None

def _GenerateFromEval(config, base, value_type):
    """Evaluate a string as the provided type
    """
    #print('Start Eval')
    #print('config = ',galsim.config.CleanConfig(config))
    if '_value' in config:
        return config['_value'], True
    elif '_fn' in config:
        #print('Using saved function')
        fn = config['_fn']
    else:
        # If the function is not already compiled, then this is the first time through, so do
        # a full parsing of all the possibilities.

        # These will be the variables to use for evaluating the eval statement.
        # Start with the current locals and globals, and add extra items to them.
        if '_eval_gdict' not in base:
            gdict = globals().copy()
            # We allow the following modules to be used in the eval string:
            exec('import galsim', gdict)
            exec('import math', gdict)
            exec('import numpy', gdict)
            exec('import numpy as np', gdict)
            exec('import os', gdict)
            exec('import astropy.units as u', gdict)
            ImportModules(base, gdict)
            base['_eval_gdict'] = gdict
        else:
            gdict = base['_eval_gdict']

        if 'str' not in config:
            raise GalSimConfigError(
                "Attribute str is required for type = %s"%(config['type']))
        string = config['str']

        # Turn any "Current" items indicated with an @ sign into regular variables.
        if '@' in string:
            # Find @items using regex.  They can include alphanumeric chars plus '.'.
            keys = re.findall(r'@[\w\.]*', string)
            #print('@keys = ',keys)
            # Remove duplicates, then sort in reverse order.
            # This makes e.g. @gal.index get processed before @gal, so replacing @gal
            # doesn't mess up the later @gal.index replacement.
            keys = sorted(np.unique(keys).tolist())
            keys.reverse()
            #print('unique @keys = ',keys)
            for key0 in keys:
                key = key0[1:] # Remove the @ sign.
                value = GetCurrentValue(key, base)
                # Give a probably unique name to this value
                key_name = "temp_variable_" + key.replace('.','_')
                #print('key_name = ',key_name)
                #print('value = ',value)
                # Replaces all occurrences of key0 with the key_name.
                string = string.replace(key0,key_name)
                # Finally, bring the key's variable name into scope.
                config['x' + key_name] = { 'type' : 'Current', 'key' : key }

        # The parameters to the function are the keys in the config dict minus their initial char.
        params = [ key[1:] for key in config.keys() if key not in eval_ignore ]

        # Also bring in any top level eval_variables that might be relevant.
        if 'eval_variables' in base:
            #print('found eval_variables = ',galsim.config.CleanConfig(base['eval_variables']))
            if not isinstance(base['eval_variables'],dict):
                raise GalSimConfigError("eval_variables must be a dict")
            for key in base['eval_variables']:
                # Only add variables that appear in the string.
                if _isWordInString(key[1:],string) and key[1:] not in params:
                    config[key] = { 'type' : 'Current',
                                    'key' : 'eval_variables.' + key }
                    params.append(key[1:])

        # Also check for the allowed base variables:
        for key in eval_base_variables:
            if key in base and _isWordInString(key,string) and key not in params:
                config['x' + key] = { 'type' : 'Current', 'key' : key }
                params.append(key)

        # Propagate index_key, rng_num as needed.
        PropagateIndexKeyRNGNum(config)
        #print('params = ',params)
        #print('config = ',config)

        # Now compile the string into a lambda function, which will be faster for subsequent
        # passes into this builder.
        try:
            if len(params) == 0:
                value = eval(string, gdict)
                config['_value'] = value
                return value, True
            else:
                fn_str = 'lambda %s: %s'%(','.join(params), string)
                #print('fn_str = ',fn_str)
                fn = eval(fn_str, gdict)
                config['_fn'] = fn
        except Exception as e:
            raise GalSimConfigError(
                "Unable to evaluate string %r as a %s\n%r"%(string, value_type, e))

    # Always need to evaluate any parameters to pass to the function
    opt = {}
    for key in config.keys():
        if key not in eval_ignore:
            opt[key] = _type_by_letter(key)
    #print('opt = ',opt)
    params, safe = GetAllParams(config, base, opt=opt, ignore=eval_ignore)
    #print('params = ',params)

    # Strip off the first character of the keys
    params = { key[1:] : value for key, value in params.items() }
    #print('params => ',params)

    # Evaluate the compiled function
    try:
        val = fn(**params)
        #print('val = ',val)
        return val, safe
    except Exception as e:
        raise GalSimConfigError(
            "Unable to evaluate string %r as a %s\n%r"%(config['str'],value_type, e))


# Register this as a valid value type
RegisterValueType('Eval', _GenerateFromEval,
                  [ float, int, bool, str, Angle, Shear, PositionD, CelestialCoord,
                    LookupTable, dict, list, None, Quantity, Unit ])
